
import os
import codecs

import kindred
import bioc
import six

def convertKindredCorpusToBioCCollection(corpus):
	assert isinstance(corpus,kindred.Corpus)
	collection = bioc.BioCCollection()
	for kdoc in corpus.documents:
		assert isinstance(kdoc,kindred.Document)

		biocDoc = bioc.BioCDocument()
		collection.add_document(biocDoc)

		if 'id' in kdoc.metadata:
			biocDoc.id = kdoc.metadata['id']
		biocDoc.infons = kdoc.metadata

		passage = bioc.BioCPassage()
		passage.text = kdoc.text
		passage.offset = 0
		biocDoc.add_passage(passage)

		seenEntityIDs = set()
		kindredID2BiocID = {}
		for e in kdoc.entities:
			assert isinstance(e,kindred.Entity)

			a = bioc.BioCAnnotation()
			a.infons = {'type':e.entityType}
			if e.sourceEntityID is None:
				a.id = str(e.entityID)
			else:
				a.id = e.sourceEntityID

			assert not a.id in seenEntityIDs
			seenEntityIDs.add(a.id)
			kindredID2BiocID[e.entityID] = a.id

			for start,end in e.position:
				l = bioc.BioCLocation(offset=start, length=(end-start))
				a.locations.append(l)

			passage.annotations.append(a)

		for r in kdoc.relations:
			assert isinstance(r,kindred.Relation)
			biocR = bioc.BioCRelation()
			biocR.infons = {'type':r.relationType}
			
			entitiesInRelation = r.entities
			argNames = r.argNames
			if argNames is None:
				argNames = [ "arg%d" % i for i,_ in enumerate(entitiesInRelation) ]

			for argName,entity in zip(argNames,entitiesInRelation):
				node = bioc.BioCNode(role=argName, refid=kindredID2BiocID[entity.entityID])
				biocR.nodes.append(node)

			passage.relations.append(biocR)

	return collection

def saveDocToSTFormat(data,txtPath,a1Path,a2Path):
	assert isinstance(data,kindred.Document)

	with codecs.open(txtPath,'w','utf8') as txtFile, codecs.open(a1Path,'w','utf8') as a1File, codecs.open(a2Path,'w','utf8') as a2File:
		txtFile.write(data.text)
		
		for e in data.entities:
			assert isinstance(e,kindred.Entity)
			assert isinstance(e.sourceEntityID,six.string_types), "Entities must have a sourceEntityID (e.g. T1) to be saved in the standoff format"
		
			positions = ";".join("%d %d" % (start,end) for start,end in e.position)
			line = "%s\t%s %s\t%s" % (e.sourceEntityID,e.entityType,positions,e.text)
			a1File.write(line+"\n")
			
		
		for i,r in enumerate(data.relations):
			assert isinstance(r,kindred.Relation)
			
			relationType = r.relationType
			relationEntityIDs = [ entity.sourceEntityID for entity in r.entities ]
			
			if r.argNames is None:
				argNames = [ ("arg%d" % (argI+1)) for argI in range(len(relationEntityIDs)) ]
			else:
				argNames = r.argNames

			arguments = " ".join(["%s:%s" % (a,b) for a,b in zip(argNames,relationEntityIDs) ])
			line = "R%d\t%s %s" % (i+1,relationType,arguments)
			a2File.write(line+"\n")

def save(corpus,dataFormat,path):
	"""
	Save a corpus to a directory
	
	:param corpus: The corpus of documents to save
	:param dataFormat: Format of data to save (only 'standoff' and 'biocxml' are supported currently)
	:param path: Path where corpus should be saved. Must be an existing directory for 'standoff'.
	:type corpus: kindred.Corpus
	:type dataFormat: str
	:type path: str
	"""
	
	assert dataFormat in ['standoff','biocxml']

	assert isinstance(corpus,kindred.Corpus)

	if dataFormat == 'standoff':
		assert os.path.isdir(path), "Path must be an existing directory"

		for i,d in enumerate(corpus.documents):
			if d.sourceFilename is None:
				base = "%08d" % i
			else:
				base = d.sourceFilename

			txtPath = os.path.join(path,'%s.txt' % base)
			a1Path = os.path.join(path,'%s.a1' % base)
			a2Path = os.path.join(path,'%s.a2' % base)

			saveDocToSTFormat(d,txtPath,a1Path,a2Path)
	elif dataFormat == 'biocxml':
		assert not os.path.isdir(path), "Path cannot be an existing directory for 'biocxml'."

		collection = convertKindredCorpusToBioCCollection(corpus)
		with bioc.iterwrite(path) as writer:
			for doc in collection.documents:
				writer.writedocument(doc)

	




import kindred
import six

class Relation:
	"""
	Describes relationship between entities (including relation type and argument names if applicable).
	
	:ivar relationType: Type of relation
	:ivar entities: List of entities in relation
	:ivar argNames: Names of relation argument associated with each entity
	:ivar probability: Optional probability for predicted relations
	"""
	
	def __init__(self,relationType=None,entities=[],argNames=None,probability=None):
		"""
		Constructor for Relation class
		
		:param relationType: Type of relation
		:param entities: List of entities in relation
		:param argNames: Names of relation argument associated with each entity
		:param probability: Optional probability for predicted relations
		:type relationType: str
		:type entities: list of kindred.Entity
		:type argNames: list of str
		:type probability: float
		"""

		assert relationType is None or isinstance(relationType, six.string_types), "relationType must be a string"
		self.relationType = relationType

		assert isinstance(entities,list),  "entities must be a list of kindred.Entity"
		for entity in entities:
			assert isinstance(entity, kindred.Entity), "entities must be a list of kindred.Entity"
		self.entities = entities

		if argNames == None:
			self.argNames = None
		else:
			assert len(argNames) == len(entities)
			self.argNames = [ str(a) for a in argNames ]

		if not probability is None:
			assert isinstance(probability, float)
		self.probability = probability
	
	def __eq__(self, other):
		"""Override the default Equals behavior"""
		if isinstance(other, self.__class__):
			return self.__dict__ == other.__dict__
		return False
	
	def __ne__(self, other):
		"""Define a non-equality test"""
		return not self.__eq__(other)

	def __str__(self):
		return "<Relation %s %s %s>" % (self.relationType,str(self.entities),str(self.argNames))

	def __repr__(self):
		return self.__str__()

	def __hash__(self):
		if self.argNames is None:
			return hash((self.relationType,tuple(self.entities),self.probability))
		else:
			return hash((self.relationType,tuple(self.entities),tuple(self.argNames),self.probability))





from sklearn import svm
from sklearn.linear_model import LogisticRegression
from collections import defaultdict

import kindred
from kindred.CandidateBuilder import CandidateBuilder
from kindred.Vectorizer import Vectorizer
import numpy as np

class RelationClassifier:
	"""
	Manages binary classifier(s) for relation classification.
	
	:param classifierType: Which classifier is used ('SVM' or 'LogisticRegression')
	:param tfidf: Whether it will use tfidf for the vectorizer
	:param features: A list of specific features. Valid features are "entityTypes", "unigramsBetweenEntities", "bigrams", "dependencyPathEdges", "dependencyPathEdgesNearEntities"
	:param threshold: A specific threshold to use for classification (which will then use a logistic regression classifier)
	:param entityCount: Number of entities in each relation (default=2). Passed to the CandidateBuilder (if needed)
	:param acceptedEntityTypes: Tuples of entity types that relations must match. None will match allow relations of any entity types. Passed to the CandidateBuilder (if needed)
	:param isTrained: Whether the classifier has been trained yet. Will throw an error if predict is called before it is trained.
	"""
	
	def __init__(self,classifierType='SVM',tfidf=True,features=None,threshold=None,entityCount=2,acceptedEntityTypes=None):
		"""
		Constructor for the RelationClassifier class
		
		:param classifierType: Which classifier to use (must be 'SVM' or 'LogisticRegression')
		:param tfidf: Whether to use tfidf for the vectorizer
		:param features: A list of specific features. Valid features are "entityTypes", "unigramsBetweenEntities", "bigrams", "dependencyPathEdges", "dependencyPathEdgesNearEntities"
		:param threshold: A specific threshold to use for classification (which will then use a logistic regression classifier)
		:param entityCount: Number of entities in each relation (default=2). Passed to the CandidateBuilder (if needed)
		:param acceptedEntityTypes: Tuples of entity types that relations must match. None will match allow relations of any entity types. Passed to the CandidateBuilder (if needed)
		:type classifierType: str
		:type tfidf: bool
		:type features: list of str
		:type threshold: float
		:type entityCount: int
		:type acceptedEntityTypes: list of tuples
		"""
		assert classifierType in ['SVM','LogisticRegression'], "classifierType must be 'SVM' or 'LogisticRegression'"
		assert classifierType == 'LogisticRegression' or threshold is None, "Threshold can only be used when classifierType is 'LogisticRegression'"

		assert isinstance(tfidf,bool)
		assert threshold is None or isinstance(threshold,float)
		assert isinstance(entityCount,int)
		assert acceptedEntityTypes is None or isinstance(acceptedEntityTypes,list)

		self.isTrained = False
		self.classifierType = classifierType
		self.tfidf = tfidf

		self.entityCount = entityCount
		self.acceptedEntityTypes = acceptedEntityTypes

		self.chosenFeatures = ["entityTypes","unigramsBetweenEntities","bigrams","dependencyPathEdges","dependencyPathEdgesNearEntities"]
		if not features is None:
			assert isinstance(features,list)
			self.chosenFeatures = features
			
		self.threshold = threshold

	def train(self,corpus):
		"""
		Trains the classifier using this corpus. All relations in the corpus will be used for training.

		:param corpus: Corpus to use for training
		:type corpus: kindred.Corpus
		"""
		assert isinstance(corpus,kindred.Corpus)

		if not corpus.parsed:
			parser = kindred.Parser()
			parser.parse(corpus)
		
		self.candidateBuilder = CandidateBuilder(entityCount=self.entityCount,acceptedEntityTypes=self.acceptedEntityTypes)
		candidateRelations = self.candidateBuilder.build(corpus)

		if len(candidateRelations) == 0:
			raise RuntimeError("No candidate relations found in corpus for training. Does the corpus contain text and entity annotations with at least one sentence containing %d entities." % (self.entityCount))

		candidateRelationKeys = set()
		for cr in candidateRelations:
			assert isinstance(cr,kindred.CandidateRelation)
			for knownType,knownArgNames in cr.knownTypesAndArgNames:
				relKey = tuple([knownType] + knownArgNames)
				candidateRelationKeys.add(relKey)
		
		# Create mappings from the class index to a relation type and back again
		self.colToRelType = sorted(list(candidateRelationKeys))
		self.relTypeToCol = { relationType:i for i,relationType in enumerate(self.colToRelType) }
		
		Y = np.zeros((len(candidateRelations),len(self.colToRelType)),np.int32)
		
		candidateClasses = []
		for i,cr in enumerate(candidateRelations):
			for knownType,knownArgNames in cr.knownTypesAndArgNames:
				relKey = tuple([knownType] + knownArgNames)
				col = self.relTypeToCol[relKey]
				Y[i,col] = 1

		entityCountsInRelations = set([ len(r.entities) for r in corpus.getRelations() ])
		entityCountsInRelations = sorted(list(set(entityCountsInRelations)))
		assert self.entityCount in entityCountsInRelations, "Relation classifier is expecting to train on relations with %d entities (entityCount=%d). But the known relations in the corpus contain relations with the following entity counts: %s. Perhaps the entityCount parameter should be changed or there is a problem with the training corpus." % (self.entityCount,self.entityCount,str(entityCountsInRelations))

		self.relTypeToValidEntityTypes = defaultdict(set)

		for d in corpus.documents:
			for r in d.relations:
				validEntityTypes = tuple([ e.entityType for e in r.entities ])
				
				relKey = tuple([r.relationType] + r.argNames)
				self.relTypeToValidEntityTypes[relKey].add(validEntityTypes)

		self.vectorizer = Vectorizer(entityCount=self.entityCount,featureChoice=self.chosenFeatures,tfidf=self.tfidf)
		trainVectors = self.vectorizer.fit_transform(candidateRelations)
	
		assert trainVectors.shape[0] == Y.shape[0]

		posCount = Y.sum()
		negCount = Y.shape[0]*Y.shape[1] - posCount

		assert negCount > 0, "Must have at least one negative candidate relation in set for training"
		assert posCount > 0, "Must have at least one positive candidate relation in set for training"

		self.clf = None
		if self.classifierType == 'SVM':
			self.clf = kindred.MultiLabelClassifier(svm.LinearSVC,class_weight='balanced',random_state=1,max_iter=10000)
		elif self.classifierType == 'LogisticRegression' and self.threshold is None:
			self.clf = kindred.MultiLabelClassifier(LogisticRegression,class_weight='balanced',random_state=1,solver='liblinear',multi_class='ovr')
		elif self.classifierType == 'LogisticRegression' and not self.threshold is None:
			self.clf = kindred.MultiLabelClassifier(kindred.LogisticRegressionWithThreshold,threshold=self.threshold)
		
		self.clf.fit(trainVectors,Y)
		
		self.isTrained = True

	def predict(self,corpus):
		"""
		Use the relation classifier to predict new relations for a corpus. The new relations will be added to the Corpus.

		:param corpus: Corpus to make predictions on
		:type corpus: kindred.Corpus
		"""
		assert self.isTrained, "Classifier must be trained using train() before predictions can be made"
	
		assert isinstance(corpus,kindred.Corpus)
		
		if not corpus.parsed:
			parser = kindred.Parser()
			parser.parse(corpus)
		
		candidateRelations = self.candidateBuilder.build(corpus)

		# Check if there are any candidate relations to classify in this corpus
		if len(candidateRelations) == 0:
			return
		
		predictedRelations = []
		testVectors = self.vectorizer.transform(candidateRelations)

		classMatrix = self.clf.predict(testVectors)
		if self.clf.has_predict_proba():
			probMatrix = self.clf.predict_proba(testVectors)
		else:
			probMatrix = None


		predictedProb = None
		for matrixRow,matrixCol in zip(*classMatrix.nonzero()):
			candidateRelation = candidateRelations[matrixRow]

			if probMatrix is not None:
				predictedProb = probMatrix[matrixRow,matrixCol]

			relKey = self.colToRelType[matrixCol]
			relType = relKey[0]
			argNames = relKey[1:]
			
			candidateRelationEntityTypes = tuple( [ e.entityType for e in candidateRelation.entities ] )
			if not tuple(candidateRelationEntityTypes) in self.relTypeToValidEntityTypes[relKey]:
				continue

			predictedRelation = kindred.Relation(relType,candidateRelation.entities,argNames=argNames,probability=predictedProb)
			predictedRelations.append(predictedRelation)

		# Add the predicted relations into the corpus
		entitiesToDoc = {}
		for i,doc in enumerate(corpus.documents):
			for e in doc.entities:
				entitiesToDoc[e] = i

		for predictedRelation in predictedRelations:
			docIDs = [ entitiesToDoc[e] for e in predictedRelation.entities ]
			docIDs = list(set(docIDs))
			assert len(docIDs) > 0, "Predicted relation contains entities that don't match any documents in corpus"
			assert len(docIDs) == 1, "Predicted relation contains entities that are spread across documents"

			docID = docIDs[0]
			if not predictedRelation in corpus.documents[docID].relations:
				corpus.documents[docID].addRelation(predictedRelation)




import kindred
import random

class Corpus:
	"""
	Collection of text documents.

	:ivar documents: List of :class:`kindred.Document`
	:ivar parsed: Boolean of whether it has been parsed yet. A :class:`kindred.parser` can parse it.
	"""
	
	def __init__(self,text=None,loadFromSimpleTag=False):
		"""
		Create an empty corpus with no documents, or quickly load one with a single document using optional SimpleTag
		
		:param text: Optional SimpleTag text to initalize a single document
		:param loadFromSimpleTag: If text is provided, whether the text parameter is in the SimpleTag format and will extract entities and relations accordingly
		:type text: String (with SimpleTag format XML)
		:type loadFromSimpleTag: bool
		"""

		self.documents = []
		if not text is None:
			doc = kindred.Document(text,loadFromSimpleTag=loadFromSimpleTag)
			self.addDocument(doc)

		self.parsed = False

	def addDocument(self,doc):
		"""
		Add a single document to the corpus
		
		:param doc: Document to add
		:type doc: kindred.Document
		"""

		assert isinstance(doc,kindred.Document)
		self.documents.append(doc)

	def clone(self):
		"""
		Clone the corpus
		
		:return: Clone of the corpus
		:rtype: kindred.Corpus
		"""

		cloned = Corpus()
		for doc in self.documents:
			cloned.addDocument(doc.clone())
		return cloned
		
	def getRelations(self):
		"""
		Get all relations in this corpus
		
		:return: List of relations
		:rtype: list
		"""

		relations = []
		for doc in self.documents:
			relations += doc.relations
		return relations

	def removeRelations(self):
		"""
		Remove all relations in this corpus
		"""

		for doc in self.documents:
			doc.removeRelations()

	def split(self,trainFraction):
		"""
		Randomly split the corpus into two corpus for use as a training and test set

		:param trainFraction: Fraction of documents to use in training set
		:type trainFraction: float
		:return: Tuple of training and test corpus
		:rtype: (kindred.Corpus,kindred.Corpus)
		"""
		assert isinstance(trainFraction,float)
		assert trainFraction > 0.0 and trainFraction < 1.0
		trainIndices = random.sample(range(len(self.documents)),int(round(trainFraction*len(self.documents))))
		trainIndices = set(trainIndices)

		trainCorpus,testCorpus = kindred.Corpus(),kindred.Corpus()
		for i,doc in enumerate(self.documents):
			if i in trainIndices:
				trainCorpus.addDocument(doc)
			else:
				testCorpus.addDocument(doc)

		return trainCorpus,testCorpus

	def nfold_split(self,folds):
		"""
		Method for splitting up the corpus multiple times and is used for an n-fold cross validation approach (as a generator). Each iteration, the training and test set for that fold are provided.

		:param folds: Number of folds to create
		:type folds: int
		:return: Tuple of training and test corpus (for iterations=folds)
		:rtype: (kindred.Corpus,kindred.Corpus)
		"""
		assert isinstance(folds,int)
		assert folds > 0

		indices = list(range(len(self.documents)))
		random.shuffle(indices)

		chunkSize = int(len(self.documents)/float(folds))
		indexChunks = [ indices[i:i+chunkSize] for i in range(0,len(self.documents),chunkSize) ]

		for f in range(folds):
			trainCorpus,testCorpus = kindred.Corpus(),kindred.Corpus()
			for i,indexChunk in enumerate(indexChunks):
				for j in indexChunk:
					if i==f:
						testCorpus.addDocument(self.documents[j])
					else:
						trainCorpus.addDocument(self.documents[j])
			yield trainCorpus,testCorpus

	def splitIntoSentences(self):
		"""
		Create a new corpus with one document for each sentence in this corpus.

		:return: Corpus with one document per sentence
		:rtype: kindred.Corpus
		"""
		assert self.parsed == True, "Corpus must be parsed before it can be split into sentences"

		sentenceCorpus = kindred.Corpus()
		for doc in self.documents:
			tempCorpus = doc.splitIntoSentences()
			sentenceCorpus.documents += tempCorpus.documents
		sentenceCorpus.parsed = True

		return sentenceCorpus




import kindred
import os
import sys

def checkForDefunctKindredFiles():
	"""
	Checks and warns about defunct files that can be deleted
	"""
	if os.path.isdir(os.path.expanduser('~/.kindred')):
		sys.stderr.write("WARNING: Defunct large files in ~/.kindred from older version. This directory can be safely removed.\n")
		sys.stderr.flush()




import sys
import codecs
import os
import json
import re
from collections import OrderedDict

from xml.dom import minidom

import kindred

import bioc
from future.utils import native

def loadEntity(filename,line,text):
	assert line[0] == 'T', "ERROR in %s. Entity input should start with a T" % filename
	split = line.strip().split('\t')
	assert len(split) == 3, "ERROR in %s" % filename
	entityID = split[0]
	typeInfo = split[1]
	tokens = split[2]
		
	textChunks = []
	typeSpacePos = typeInfo.index(' ')
	typeName = typeInfo[:typeSpacePos]
	positionText = typeInfo[typeSpacePos:]
	positions = []
	for coordinates in positionText.strip().split(';'):
		a,b = coordinates.strip().split(' ')
		a,b = int(a.strip()),int(b.strip())
		textChunk = text[a:b].replace('\n',' ').strip()
		textChunks.append(textChunk)
		positions.append((a,b))
		
	# Check that the tokens match up to the text
	chunkTest = " ".join(textChunks)
	tokensTest = tokens
	chunkTest = re.sub(r'\s\s+', ' ', chunkTest)
	tokensTest = re.sub(r'\s\s+', ' ', tokensTest)
	chunkTest = chunkTest.strip()
	tokensTest = tokensTest.strip()

	assert chunkTest == tokensTest , u"ERROR in " + filename + u"For id=" + entityID + ", tokens '" + tokens.encode('ascii', 'ignore') + "' don't match up with positions: " + str(positions)
	
	entity = kindred.Entity(typeName, tokensTest, positions, entityID)

	return entity
	
def loadRelation(filename,line,ignoreComplexRelations=True):
	assert line[0] == 'E' or line[0] == 'R', "ERROR in %s. Relation input should start with a E or R" % filename
	assert ignoreComplexRelations == True, "ERROR in %s. ignoreComplexRelations must be True as kindred doesn't currently support complex relations" % filename

	split = line.strip().split('\t')
	eventInfo = split[1]
	typeSpacePos = eventInfo.index(' ')
	
	eventNameSplit = eventInfo[:typeSpacePos].split(':')
	assert len(eventNameSplit) == 1, "ERROR in %s. Cannot load trigger events" % filename
	relationType = eventNameSplit[0]
		
	isComplexRelation = False
	argumentText = eventInfo[typeSpacePos:]
	arguments = []
	for argument in argumentText.strip().split(' '):
		split2 = argument.strip().split(':')
		assert len(split2) == 2, "ERROR in %s" % filename
		tmpArgName = split2[0]
		tmpEntityID = split2[1]

		isComplexRelation = (tmpEntityID[0] == 'R' or tmpEntityID[0] == 'E')

		# We'll skip this relation as
		if ignoreComplexRelations and isComplexRelation:
			break

		assert not isComplexRelation, "ERROR in %s. kindred does not support complex relations (where one relation has another relation as an argument), use ignoreComplexRelations=True to ignore these" % filename

		assert not tmpArgName in arguments, "ERROR in %s" % filename
		arguments.append((tmpArgName,tmpEntityID))

	if ignoreComplexRelations and isComplexRelation:
		return None

	arguments = sorted(arguments)
	sourceEntityIDs = [ entityID for argName,entityID in arguments ]
	argNames = [ argName for argName,entityID in arguments ]

	relationTuple = (relationType,sourceEntityIDs,argNames)
	return relationTuple
	
# TODO: Deal with complex relations more clearly
def loadDataFromStandoff(txtFile,ignoreEntities=[],ignoreComplexRelations=True):
	annotationExtensions = ['ann','a1','a2']
	assert ignoreComplexRelations == True, "ignoreComplexRelations must be True as kindred doesn't currently support complex relations"

	with codecs.open(txtFile, "r", "utf-8") as f:
		text = f.read()

	assert txtFile.endswith('.txt')
	base = txtFile[:-4]

	annotationFiles = [ "%s.%s" % (base,ext) for ext in annotationExtensions ]
	annotationFiles = [ filename for filename in annotationFiles if os.path.isfile(filename) ]

	entities = []

	for annotationFile in annotationFiles:
		with codecs.open(annotationFile, "r", "utf-8") as f:
			for line in f:
				if line.startswith('T'):
					entity = loadEntity(annotationFile,line.strip(), text)
					if (not entity is None) and (not entity.entityType in ignoreEntities):
						entities.append(entity)
		
	sourceEntityIDToEntity = { entity.sourceEntityID:entity for entity in entities }

	relations = []
	for annotationFile in annotationFiles:
		with codecs.open(annotationFile, "r", "utf-8") as f:
			for line in f:
				if line.startswith('E') or line.startswith('R'):
					relationTuple = loadRelation(annotationFile,line.strip(),ignoreComplexRelations)
					if not relationTuple is None:
						relationType,sourceEntityIDs,argNames = relationTuple
						for sourceEntityID in sourceEntityIDs:
							assert sourceEntityID in sourceEntityIDToEntity, "Relation exists that references a non-existent entity (%s) associated with %s" % (sourceEntityID,txtFile)
						entitiesInRelation = [ sourceEntityIDToEntity[sourceEntityID] for sourceEntityID in sourceEntityIDs ]
						relation = kindred.Relation(relationType,entitiesInRelation,argNames)
						relations.append(relation)

	baseTxtFile = os.path.basename(txtFile)
	baseFilename = baseTxtFile[0:-4]

	combinedData = kindred.Document(text,entities=entities,relations=relations,sourceFilename=baseFilename)
			
	return combinedData

def parseJSON(data,ignoreEntities=[]):
	entities = []
	relations = []

	if isinstance(data,list):
		assert len(data) == 1 and isinstance(data[0],dict), "JSON loading expects a dictionary or a list with one dictionary in it"
		data = data[0]
	assert isinstance(data,dict), "JSON loading expects a dictionary or a list with one dictionary in it"

	text = data['text']
	if 'denotations' in data:
		for d in data['denotations']:
			sourceEntityID = None
			if 'id' in d:
				sourceEntityID = d['id']
			
			entityType = d['obj']
			span = d['span']
			startPos,endPos = span['begin'],span['end']
			position = [(startPos,endPos)]
			entityText = text[startPos:endPos]
			
			if not entityType in ignoreEntities:
				entity = kindred.Entity(entityType,entityText,position,sourceEntityID=sourceEntityID)
				entities.append(entity)

	sourceEntityIDToEntity = { entity.sourceEntityID:entity for entity in entities }

	if 'relations' in data:
		for r in data['relations']:
			obj = r['obj']
			relationType = r['pred']
			subj = r['subj']
			
			sourceEntityIDs = [obj,subj]
			argNames = ['obj','subj']
			entitiesInRelation = [ sourceEntityIDToEntity[sourceEntityID] for sourceEntityID in sourceEntityIDs ]
		
			relation = kindred.Relation(relationType,entitiesInRelation,argNames)
			relations.append(relation)
	
	expected = ['denotations','divid','modifications','namespaces','project','relations','sourcedb','sourceid','target','text','tracks']
	extraFields = [ k for k in data.keys() if not k in expected]
	assert len(extraFields) == 0, "Found additional unexpected fields (%s) in JSON" % (",".join(extraFields))
		
	combinedData = kindred.Document(text,entities=entities,relations=relations)

	return combinedData

def loadDataFromJSON(filename,ignoreEntities=[]):
	with open(filename) as f:
		data = json.load(f)
	parsed = parseJSON(data,ignoreEntities)
	
	baseTxtFile = os.path.basename(filename)
	parsed.sourceFilename = baseTxtFile
	
	return parsed
	
def parseSimpleTag_helper(node,currentPosition=0,ignoreEntities=[]):
	text,entities,relationTuples = '',[],[]
	for s in node.childNodes:
		if s.nodeType == s.ELEMENT_NODE:
			insideText,insideEntities,insideRelationTuples = parseSimpleTag_helper(s,currentPosition+len(text))

			if s.tagName == 'relation':
				relationType = s.getAttribute('type')
				arguments = [ (argName,entityID) for argName,entityID in s.attributes.items() if argName != 'type' ]
				arguments = sorted(arguments)
				
				sourceEntityIDs = [ sourceEntityID for argName,sourceEntityID in arguments]
				argNames = [ argName for argName,sourceEntityID in arguments]
				

				relationTuple = (relationType,sourceEntityIDs,argNames)
				relationTuples.append(relationTuple)
			else: # Entity
				entityType = s.tagName
				sourceEntityID = s.getAttribute('id')
				position = [(currentPosition+len(text),currentPosition+len(text)+len(insideText))]

				assert len(insideText) > 0, "Name (text inside tags) is empty for entity of type %s" % entityType

				if not entityType in ignoreEntities:
					e = kindred.Entity(entityType,insideText,position,sourceEntityID=sourceEntityID)
					entities.append(e)
				
			text += insideText
			entities += insideEntities
			relationTuples += insideRelationTuples
		elif s.nodeType == s.TEXT_NODE:
			text += s.nodeValue
			
	return text,entities,relationTuples
	
def mergeEntitiesWithMatchingIDs(unmergedEntities):
	assert isinstance(unmergedEntities,list)

	entityDict = OrderedDict()
	for e in unmergedEntities:
		assert isinstance(e, kindred.Entity)
		if e.sourceEntityID in entityDict:
			entityDict[e.sourceEntityID].text += " " + e.text
			entityDict[e.sourceEntityID].position += e.position
		else:
			entityDict[e.sourceEntityID] = e
			
	return list(entityDict.values())
	
def parseSimpleTag(text,ignoreEntities=[]):
	docText = u"<doc>%s</doc>" % text
	xmldoc = minidom.parseString(docText.encode('utf8'))
	docNode = xmldoc.childNodes[0]
	text,unmergedEntities,relationTuples = parseSimpleTag_helper(docNode,ignoreEntities=ignoreEntities)
	
	missingSourceEntityID = [ e.sourceEntityID == '' for e in unmergedEntities ]
	assert all(missingSourceEntityID) or (not any(missingSourceEntityID)), 'All entities or none (not some) should be given IDs'
	assert (not any(missingSourceEntityID)) or len(relationTuples) == 0, "Cannot include relations with no-ID entities"
	
	if all(missingSourceEntityID):
		for i,e in enumerate(unmergedEntities):
			e.sourceEntityID = i+1
					
	entities = mergeEntitiesWithMatchingIDs(unmergedEntities)
		
	sourceEntityIDToEntity = { entity.sourceEntityID:entity for entity in entities }

	relations = []
	for relationType,sourceEntityIDs,argNames in relationTuples:
		assert len(sourceEntityIDs) == len(argNames)
		entitiesInRelation = [ sourceEntityIDToEntity[sourceEntityID] for sourceEntityID in sourceEntityIDs ]
		relation = kindred.Relation(relationType=relationType,entities=entitiesInRelation,argNames=argNames)
		relations.append(relation)

	combinedData = kindred.Document(text,entities=entities,relations=relations)
	return combinedData

def convertBiocDocToKindredDocs(document):
	assert isinstance(document,bioc.BioCDocument)
	kindredDocs = []
	for passage in document.passages:
		assert isinstance(passage,bioc.BioCPassage)
		
		text = passage.text
		offset = int(native(passage.offset))
		entities = []
		relations = []
		
		for a in passage.annotations:
			assert isinstance(a,bioc.BioCAnnotation)
			
			entityType = a.infons['type']
			sourceEntityID = a.id
			
			position = []
			segments = []
			
			for l in a.locations:
				assert isinstance(l,bioc.BioCLocation)
				startPos = int(native(l.offset)) - offset
				endPos = startPos + int(native(l.length))
				position.append((startPos,endPos))
				segments.append(text[startPos:endPos])
			
			entityText = " ".join(segments)
			e = kindred.Entity(entityType,entityText,position,sourceEntityID)
			entities.append(e)

		sourceEntityIDToEntity = { entity.sourceEntityID:entity for entity in entities }
			
		for r in passage.relations:
			assert isinstance(r,bioc.BioCRelation)
			relationType = r.infons['type']
			
			arguments = []
			for n in r.nodes:
				assert isinstance(n,bioc.BioCNode)
				arguments.append((n.role,n.refid))
			arguments = sorted(arguments)
				
			sourceEntityIDs = [ entityID for argName,entityID in arguments]
			argNames = [ argName for argName,entityID in arguments]
			entities = [ sourceEntityIDToEntity[sourceEntityID] for sourceEntityID in sourceEntityIDs ]
			
			r = kindred.Relation(relationType,entities,argNames)
			relations.append(r)
		
		metadata = dict(document.infons)
		metadata.update(passage.infons)
		metadata['id'] = document.id
		relData = kindred.Document(text,entities=entities,relations=relations,metadata=metadata)
		kindredDocs.append(relData)

	return kindredDocs

def loadDataFromBioC(filename,ignoreEntities=[]):
	with open(filename, 'r') as fp:
		collection = bioc.load(fp)
	assert isinstance(collection,bioc.BioCCollection)
	
	parsed = []
	for document in collection.documents:
		parsed += convertBiocDocToKindredDocs(document)

	corpus = kindred.Corpus()
	corpus.documents = parsed
	return corpus

def iterLoad(dataFormat,path,corpusSizeCutoff=500):
	"""
	Iteratively load sections of a (presumably large) corpus. This will create a generator that provides kindred.Corpus objects that are subsets of the larger corpus. This should be used to lower the memory requirements (so that the entire file doesn't need to be loaded into memory at one time).

	:param dataFormat: Format of the data files to load (only 'biocxml' is currently supported)
	:param path: Path to data. Can be directory or an individual file (for bioc, json or simpletag)
	:param corpusSizeCutoff: Approximate maximum number of documents to be in each corpus subset
	:type dataFormat: str
	:type path: str
	:type corpusSizeCutoff: int
	:return: Subsets of the BioC file
	:rtype: A kindred.Corpus generator
	"""
	assert dataFormat == 'biocxml'

	corpus = kindred.Corpus()

	if os.path.isdir(path):
		filenames = [ os.path.join(path,x) for x in os.listdir(path) if x.endswith('bioc.xml') ]
	else:
		filenames = [path]

	for filename in filenames:
		with bioc.iterparse(filename) as parser:
			for document in parser:
				if len(corpus.documents) >= corpusSizeCutoff:
					yield corpus
					corpus = kindred.Corpus()
				kindredDocs = convertBiocDocToKindredDocs(document)
				for kindredDoc in kindredDocs:
					corpus.addDocument(kindredDoc)

	if len(corpus.documents) > 0:
		yield corpus
	
def xor(a, b):
	return (a and not b) or (not a and b)

def load(dataFormat,path,ignoreEntities=[],ignoreComplexRelations=True):
	"""
	Load a corpus from a variety of formats. If path is a directory, it will try to load all files of the corresponding data type. For standoff format, it will use any associated annotations files (with suffixes .ann, .a1 or .a2)
	
	:param dataFormat: Format of the data files to load ('standoff','biocxml','json','simpletag')
	:param path: Path to data. Can be directory or an individual file. Should be the txt file for standoff.
	:param ignoreEntities: List of entity types to ignore while loading
	:param ignoreComplexRelations: Whether to filter out relations where one argument is another relation (must be True as kindred doesn't currently support complex relations)
	:type dataFormat: str
	:type path: str
	:type ignoreEntities: list
	:type ignoreComplexRelations: bool
	:return: Corpus of loaded documents
	:rtype: kindred.Corpus
	"""
	assert dataFormat in ['standoff','biocxml','json','simpletag']
	assert ignoreComplexRelations == True, "ignoreComplexRelations must be True as kindred doesn't currently support complex relations"

	corpus = kindred.Corpus()

	if os.path.isdir(path):
		directory = path

		filenames = sorted(list(os.listdir(directory)))

		for filename in filenames:
			if dataFormat == 'standoff' and filename.endswith('.txt'):
				absPath = os.path.join(directory, filename)
				doc = loadDataFromStandoff(absPath,ignoreEntities=ignoreEntities)
				corpus.addDocument(doc)
			elif dataFormat == 'biocxml' and filename.endswith('.bioc.xml'):
				absPath = os.path.join(directory, filename)
				tempCorpus = loadDataFromBioC(absPath,ignoreEntities=ignoreEntities)
				corpus.documents += tempCorpus.documents
			elif dataFormat == 'json' and filename.endswith('.json'):
				absPath = os.path.join(directory, filename)
				doc = loadDataFromJSON(absPath,ignoreEntities=ignoreEntities)
				corpus.addDocument(doc)
			elif dataFormat == 'simpletag' and filename.endswith('.simple'):
				absPath = os.path.join(directory, filename)
				with open(absPath,'r') as f:
					filecontents = f.read().strip()

				doc = parseSimpleTag(filecontents,ignoreEntities=ignoreEntities)
				doc.sourceFilename = filename
				corpus.addDocument(doc)

	elif dataFormat == 'standoff':
		doc = loadDataFromStandoff(path,ignoreEntities=ignoreEntities)
		corpus.addDocument(doc)
	elif dataFormat == 'biocxml':
		corpus = loadDataFromBioC(path,ignoreEntities=ignoreEntities)
	elif dataFormat == 'json':
		doc = loadDataFromJSON(path,ignoreEntities=ignoreEntities)
		corpus.addDocument(doc)
	elif dataFormat == 'simpletag':
		with open(path,'r') as f:
			filecontents = f.read().strip()

		doc = parseSimpleTag(filecontents,ignoreEntities=ignoreEntities)
		doc.sourceFilename = os.path.basename(path)
		corpus.addDocument(doc)
	
	if len(corpus.documents) == 0:
		raise RuntimeError("No documents loaded from directory (%s). Are you sure this directory contains the corpus (format: %s)" % (directory,dataFormat))

	return corpus
			





from collections import defaultdict
import itertools

import kindred

class CandidateBuilder:
	"""
	Generates set of all possible relations in corpus.
	
	:ivar entityCount: Number of entities in each relation (default=2)
	:ivar acceptedEntityTypes: Tuples of entities that candidate relations must match. Each entity should be the same length as entityCount. None will match all candidate relations.
	"""
	def __init__(self,entityCount=2,acceptedEntityTypes=None):
		"""
		Constructor

		:param entityCount: Number of entities in each relation (default=2)
		:param acceptedEntityTypes: Tuples of entities that candidate relations must match. Each entity should be the same length as entityCount. None will match all candidate relations.
		:type entityCount: int
		:type acceptedEntityTypes: list of tuples
		"""
		assert isinstance(entityCount,int)
		assert entityCount >= 2
		self.entityCount = entityCount

		assert acceptedEntityTypes is None or isinstance(acceptedEntityTypes,list)
		if acceptedEntityTypes is None:
			self.acceptedEntityTypes = None
		else:
			for acceptedEntityType in acceptedEntityTypes:
				assert isinstance(acceptedEntityType,tuple)
				assert len(acceptedEntityType) == entityCount
			self.acceptedEntityTypes = set(acceptedEntityTypes)

	def build(self,corpus):
		"""
		Creates the set of all possible relations that exist within the given corpus. Each relation will be contained within a single sentence.
		
		:param corpus: Corpus of text with which to build relation candidates
		:type corpus: kindred.Corpus
		:return: List of candidate relations matching entityCount and acceptedEntityTypes
		:rtype: List of kindred.Relation
		"""
		assert isinstance(corpus,kindred.Corpus)
		assert corpus.parsed, "Corpus must have already been parsed"

		candidates = []
		for doc in corpus.documents:
			existingRelationsAndArgNames = defaultdict(list)
			for r in doc.relations:
				assert isinstance(r,kindred.Relation)
				entities = tuple(r.entities)
				existingRelationsAndArgNames[entities].append((r.relationType,tuple(r.argNames)))

			for sentence in doc.sentences:
				entitiesInSentence = [ entity for entity,tokenIndices in sentence.entityAnnotations ]
							
				for entitiesInRelation in itertools.permutations(entitiesInSentence, self.entityCount):
					typesInRelation = tuple([ e.entityType for e in entitiesInRelation ])
					if not self.acceptedEntityTypes is None and not typesInRelation in self.acceptedEntityTypes:
						# Relation doesn't contain the right entity types (so skip it)
						continue

					knownTypesAndArgNames = list(set(existingRelationsAndArgNames[entitiesInRelation]))
					knownTypesAndArgNames = [ (relationType,list(argNames)) for relationType,argNames in knownTypesAndArgNames ]
				
					candidateRelation = kindred.CandidateRelation(entities=list(entitiesInRelation),knownTypesAndArgNames=knownTypesAndArgNames,sentence=sentence)
					candidates.append(candidateRelation)

		return candidates




"""
Importer for BioNLP Shared Task data
"""

import os
import tempfile
import sys
import shutil
import six

import kindred

# Load the task information into this data structure
taskOptions = {}
bionlpstFile = os.path.join(os.path.dirname(__file__),'bionlpst_files.txt')
with open(bionlpstFile,'r') as f:
	for line in f:
		if line.strip() != '':
			taskName,url,expectedFile,expectedSHA256 = line.strip().split('\t')
			taskOptions[taskName] = (url,expectedFile,expectedSHA256)

def listTasks():
	"""
	List the names of the BioNLP Shared Task datasets that can be loaded. These values can be passed to the kindred.bionlpst.load function as the taskName argument

	:return: List of valid taskNames
	:rtype: str
	"""
	
	return sorted(list(taskOptions.keys()))

def load(taskName,ignoreEntities=[]):
	"""
	Download and load the corresponding corpus from the BioNLP Shared Task
	
	:param taskName: The name of the shared task to download (e.g. 'BioNLP-ST-2016_BB-event_train'). Use kindred.bionlpst.listTasks() to get a list of valid options
	:param ignoreEntities: A list of any entities that should be ignored during loading
	:type taskName: str
	:type ignoreEntities: list of str
	:return: The loaded corpus
	:rtype: kindred.Corpus
	"""
	global taskOptions

	tempDir = tempfile.mkdtemp()

	assert taskName in taskOptions.keys(), "%s not a valid option in %s" % (taskName, taskOptions.keys())
	url,expectedFile,expectedSHA256 = taskOptions[taskName]
	filesToDownload = [(url,expectedFile,expectedSHA256)]
	expectedDir = expectedFile.replace('.zip','')

	try:
		kindred.utils._downloadFiles(filesToDownload,tempDir)
	except:
		exc_info = sys.exc_info()
		shutil.rmtree(tempDir)
		six.reraise(*exc_info)

	mainDir = kindred.utils._findDir(expectedDir,tempDir)

	corpus = kindred.load(dataFormat='standoff',path=mainDir,ignoreEntities=ignoreEntities)

	shutil.rmtree(tempDir)

	return corpus




# -*- coding: utf-8 -*-


import kindred
from intervaltree import IntervalTree
from collections import defaultdict
import six

class Parser:
	"""
	Runs Spacy on corpus to get sentences and associated tokens
	
	:ivar language: Language to parse (en/de/es/pt/fr/it/nl)
	:ivar nlp: The underlying Spacy language model to use for parsing
	"""

	_languageModels = {}
	
	def __init__(self,language='en'):
		"""
		Create a Parser object that will use Spacy for parsing. It offers all the same languages that Spacy offers. Check out: https://spacy.io/usage/models. Note that the language model needs to be downloaded first (e.g. python -m spacy download en)
		
		:param language: Language to parse (en/de/es/pt/fr/it/nl)
		:type language: str
		"""

		# We only load spacy if a Parser is created (to allow ReadTheDocs to build the documentation easily)
		import spacy

		acceptedLanguages = ['en','de','es','pt','fr','it','nl']
		assert language in acceptedLanguages, "Language for parser (%s) not in accepted languages: %s" % (language,str(acceptedLanguages))

		self.language = language

		if not language in Parser._languageModels:
			Parser._languageModels[language] = spacy.load(language, disable=['ner'])

		self.nlp = Parser._languageModels[language]

	def _sentencesGenerator(self,text):
		if six.PY2 and isinstance(text,str):
			text = unicode(text)

		parsed = self.nlp(text)
		sentence = None
		for token in parsed:
			if sentence is None or token.is_sent_start:
				if not sentence is None:
					yield sentence
				sentence = []
			sentence.append(token)

		if not sentence is None and len(sentence) > 0:
			yield sentence

	def parse(self,corpus):
		"""
		Parse the corpus. Each document will be split into sentences which are then tokenized and parsed for their dependency graph. All parsed information is stored within the corpus object.
		
		:param corpus: Corpus to parse
		:type corpus: kindred.Corpus
		"""

		assert isinstance(corpus,kindred.Corpus)

		# Ignore DeprecationWarning from SortedDict which is inside IntervalTree
		import warnings
		warnings.filterwarnings("ignore", category=DeprecationWarning)

		for d in corpus.documents:
			entityIDsToEntities = { entity.entityID:entity for entity in d.entities }
		
			denotationTree = IntervalTree()
			entityTypeLookup = {}
			for e in d.entities:
				entityTypeLookup[e.entityID] = e.entityType
			
				for a,b in e.position:
					if b > a:
						denotationTree[a:b] = e.entityID
				
			for sentence in self._sentencesGenerator(d.text):
				tokens = []
				for t in sentence:
					token = kindred.Token(t.text,t.lemma_,t.pos_,t.idx,t.idx+len(t.text))
					tokens.append(token)

				sentenceStart = tokens[0].startPos
				sentenceEnd = tokens[-1].endPos
				sentenceTxt = d.text[sentenceStart:sentenceEnd]

				indexOffset = sentence[0].i
				dependencies = []
				for t in sentence:
					depName = t.dep_
					dep = (t.head.i-indexOffset,t.i-indexOffset,depName)
					dependencies.append(dep)

				entityIDsToTokenLocs = defaultdict(list)
				for i,t in enumerate(tokens):
					entitiesOverlappingWithToken = denotationTree[t.startPos:t.endPos]
					for interval in entitiesOverlappingWithToken:
						entityID = interval.data
						entityIDsToTokenLocs[entityID].append(i)

				sentence = kindred.Sentence(sentenceTxt, tokens, dependencies, d.sourceFilename)
				
				# Let's gather up the information about the "known" entities in the sentence
				for entityID,entityLocs in sorted(entityIDsToTokenLocs.items()):
					# Get the entity associated with this ID
					e = entityIDsToEntities[entityID]
					sentence.addEntityAnnotation(e,entityLocs)
					
				d.addSentence(sentence)

		corpus.parsed = True




import kindred
from collections import OrderedDict

class Document:
	"""
	Span of text with associated tagged entities and relations between entities.

	:ivar text: Text in document (plain text or SimpleTag)
	:ivar entities: Entities in document
	:ivar relations: Relations in document
	:ivar sourceFilename: Filename that this document came from
	:ivar metadata: IDs and other information associated with the source (e.g. PMID)
	:ivar sentences: List of sentences (:class:`kindred.Sentence`) if the document has been parsed
	"""
	
	def __init__(self,text,entities=None,relations=None,sourceFilename=None,metadata={},loadFromSimpleTag=False):
		"""
		Constructor for a Document that can take text using the SimpleTag XML format, or a set of Entities and Relations with associated text.
		
		:param text: Text in document (plain text or SimpleTag)
		:param entities: Entities in document
		:param relations: Relations in document
		:param sourceFilename: Filename that this document came from
		:param metadata: IDs and other information associated with the source (e.g. PMID)
		:param loadFromSimpleTag: Assumes the text parameter is in the SimpleTag format and will extract entities and relations accordingly
		:type text: str
		:type entities: list of kindred.Entity
		:type relations: list of kindred.Relation
		:type sourceFilename: str
		:type metadata: dict
		:type loadFromSimpleTag: bool
		"""

		self.sourceFilename = sourceFilename
		self.metadata = metadata

		if loadFromSimpleTag:
			assert entities is None and relations is None, 'Entities and relations will be extracted from SimpleTag. They cannot also be passed in as parameters'

			docToCopy = kindred.loadFunctions.parseSimpleTag(text)
			assert isinstance(docToCopy,kindred.Document)
			self.text = docToCopy.text
			self.entities = docToCopy.entities
			self.relations = docToCopy.relations
		else:
			self.text = text
			
			if entities is None:
				self.entities = []
			else:
				assert isinstance(entities,list)
				for e in entities:
					assert isinstance(e,kindred.Entity)
				self.entities = entities
			
			if relations is None:
				self.relations = []
			else:
				assert isinstance(relations,list)
				for r in relations:
					assert isinstance(r,kindred.Relation)
				self.relations = relations

		self.sentences = []
		
	def __repr__(self):
		"""
		String representation of Document
		
		:return: string representation
		:rtype: str
		"""
		return self.__str__()
	
	def __str__(self):
		"""
		String representation of Document
		
		:return: string representation
		:rtype: str
		"""

		return u"<Document %s %s %s>"  % (self.text,str(self.entities),str(self.relations))
	
	def addEntity(self,entity):
		"""
		Add an entity to this document
		
		:param entity: Entity to add
		:type entity: kindred.Entity
		"""

		self.entities.append(entity)

	def addRelation(self,relation):
		"""
		Add a relation to this document
		
		:param relation: Relation to add
		:type relation: kindred.Relation
		"""

		self.relations.append(relation)

	def addSentence(self,sentence):
		"""
		Add a sentence to this document
		
		:param sentence: Sentence to add
		:type sentence: kindred.Sentence
		"""

		assert isinstance(sentence,kindred.Sentence)
		self.sentences.append(sentence)
		
	def clone(self):
		"""
		Clones the document
		
		:return: Clone of the document
		:rtype: kindred.Document
		"""

		cloned = Document(self.text,entities=self.entities,relations=self.relations,sourceFilename=self.sourceFilename)
		return cloned

	def removeRelations(self):
		"""
		Remove all relations in this document
		"""
		self.relations = []


	def splitIntoSentences(self):
		"""
		Create a new corpus with one document for each sentence in this document.

		:return: Corpus with one document per sentence
		:rtype: kindred.Corpus
		"""

		sentenceCorpus = kindred.Corpus()
		
		for sentence in self.sentences:
			sentenceStart = sentence.tokens[0].startPos
			
			entitiesInSentence = [ entity for entity,tokenIndices in sentence.entityAnnotations ]

			entityMap = OrderedDict()
			for e in entitiesInSentence:
				startPos,endPos = e.position[0]
				newPosition = [ (startPos-sentenceStart, endPos-sentenceStart) ]
				newE = kindred.Entity(e.entityType,e.text,newPosition,e.sourceEntityID,e.externalID)
				entityMap[e] = newE

			relationsInSentence = [ r for r in self.relations if all( e in entitiesInSentence for e in r.entities ) ]
			newRelationsInSentence = []
			for r in relationsInSentence:
				newEntitiesInRelation = [ entityMap[e] for e in r.entities ]
				newRelation = kindred.Relation(r.relationType,newEntitiesInRelation,r.argNames,r.probability)
				newRelationsInSentence.append(newRelation)

			newEntitiesInSentence = list(entityMap.values())
			doc = kindred.Document(sentence.text,newEntitiesInSentence,newRelationsInSentence)

			newTokens = [ kindred.Token(t.word,t.lemma,t.partofspeech,t.startPos-sentenceStart,t.endPos-sentenceStart) for t in sentence.tokens ]

			newSentence = kindred.Sentence(sentence.text,newTokens,sentence.dependencies,sentence.sourceFilename)
			newEntityAnnotations = [ (entityMap[e],tokenIndices) for e,tokenIndices in sentence.entityAnnotations ]
			newSentence.entityAnnotations = newEntityAnnotations
			doc.sentences = [newSentence]

			sentenceCorpus.addDocument(doc)

		return sentenceCorpus

		




from collections import Counter
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from scipy.sparse import hstack
import itertools

import kindred

def _doEntityTypes(candidates,entityCount):
	data = []
	for cr in candidates:
		assert isinstance(cr,kindred.CandidateRelation)
		
		tokenInfo = {}
		for argI,entity in enumerate(cr.entities):
			argName = "selectedtokentypes_%d_%s" % (argI,entity.entityType)
			tokenInfo[argName] = 1
		data.append(tokenInfo)
	return data

def _doUnigramsBetweenEntities(candidates,entityCount):
	data = []	
	for cr in candidates:
		assert isinstance(cr,kindred.CandidateRelation)
		
		sentence = cr.sentence
		dataForThisCR = Counter()
		entityCount = len(cr.entities)
		entityToTokenIndices = { e:tokenIndices for e,tokenIndices in cr.sentence.entityAnnotations }
		
		for e1,e2 in itertools.combinations(range(entityCount),2):
			pos1 = entityToTokenIndices[cr.entities[e1]]
			pos2 = entityToTokenIndices[cr.entities[e2]]
			
			if max(pos1) < min(pos2):
				startPos,endPos = max(pos1)+1,min(pos2)
			else:
				startPos,endPos = max(pos2)+1,min(pos1)

			tokenData = [ sentence.tokens[i].word.lower() for i in range(startPos,endPos) ]

			basename = u"ngrams_betweenentities"
			if entityCount > 2:
				basename = u"ngrams_betweenentities_%d_%d" % (e1,e2)

			for t in tokenData:
				dataForThisCR[u"%s_%s" % (basename,t)] += 1
		data.append(dataForThisCR)

	return data

def _doDependencyPathEdges(candidates,entityCount):
	data = []	
	for cr in candidates:
		assert isinstance(cr,kindred.CandidateRelation)
		sentence = cr.sentence
		entityToTokenIndices = { e:tokenIndices for e,tokenIndices in cr.sentence.entityAnnotations }
		
		entityCount = len(cr.entities)
		dataForThisCR = Counter()
		for e1,e2 in itertools.combinations(range(entityCount),2):
			pos1 = entityToTokenIndices[cr.entities[e1]]
			pos2 = entityToTokenIndices[cr.entities[e2]]

			combinedPos = pos1 + pos2
			
			basename = u"dependencypathelements"
			if entityCount > 2:
				basename = u"dependencypathelements_%d_%d" % (e1,e2)

			nodes,edges = sentence.extractMinSubgraphContainingNodes(combinedPos)
			for a,b,dependencyType in edges:
				dataForThisCR[u"%s_%s" % (basename,dependencyType)] += 1
		data.append(dataForThisCR)

	return data

def _doDependencyPathEdgesNearEntities(candidates,entityCount):
	data = []	
	for cr in candidates:
		assert isinstance(cr,kindred.CandidateRelation)
		sentence = cr.sentence
		entityToTokenIndices = { e:tokenIndices for e,tokenIndices in cr.sentence.entityAnnotations }
		
		dataForThisCR = Counter()

		allEntityLocs = []
		for e in cr.entities:
			allEntityLocs += entityToTokenIndices[e]
		
		nodes,edges = sentence.extractMinSubgraphContainingNodes(allEntityLocs)
		for i,e in enumerate(cr.entities):
			pos = entityToTokenIndices[e]

			for a,b,dependencyType in edges:
				if a in pos:
					dataForThisCR[u"dependencypathnearselectedtoken_%d_%s" % (i,dependencyType)] += 1
		data.append(dataForThisCR)

	return data

def _doBigrams(candidates,entityCount):
	data = []
	for cr in candidates:
		assert isinstance(cr,kindred.CandidateRelation)
		
		sentence = cr.sentence
		dataForThisCR = Counter()

		for _ in cr.entities:
			startPos = 0
			endPos = len(sentence.tokens)

			tokenData = [ (sentence.tokens[i].word.lower(),sentence.tokens[i+1].word.lower()) for i in range(startPos,endPos-1) ]
			for t in tokenData:
				dataForThisCR[u"bigrams_%s_%s" % t] += 1
		data.append(dataForThisCR)

	return data

class Vectorizer:
	"""
	Vectorizes set of candidate relations into scipy sparse matrix.
	
	:ivar entityCount: Number of entities in candidate relations to vectorize
	:ivar featureChoice: List of features (can be one or a set of the following: 'entityTypes', 'unigramsBetweenEntities', 'bigrams', 'dependencyPathEdges', 'dependencyPathEdgesNearEntities'). Set as None to use all of them. 
	:ivar tfidf: Whether it will normalize n-gram based features using term frequency-inverse document frequency
	:ivar fitted: Whether it has been fit on data first (before transforming).
	:ivar dictVectorizers: Dictionary vectorizers used for each feature
	:ivar tfidfTransformers: TFIDF transformers used for each feature (if appropriate and selected)
	"""
	
	def __init__(self,entityCount=2,featureChoice=None,tfidf=True):
		"""
		Constructor for vectorizer class with options for what features to use and whether to normalize using TFIDF
		
		:param entityCount: Number of entities in candidate relations to vectorize
		:param featureChoice: List of features (can be one or a set of the following: 'entityTypes', 'unigramsBetweenEntities', 'bigrams', 'dependencyPathEdges', 'dependencyPathEdgesNearEntities'). Set as None to use all of them. 
		:param tfidf: Whether to normalize n-gram based features using term frequency-inverse document frequency
		:type entityCount: int
		:type featureChoice: list of str
		:type tfidf: bool
		"""
		
		self.fitted = False

		assert isinstance(entityCount, int)
		self.entityCount = entityCount
		
		self._registerFunctions()
		validFeatures = self.featureInfo.keys()

		if featureChoice is None:
			self.chosenFeatures = ['entityTypes','unigramsBetweenEntities','bigrams','dependencyPathEdges','dependencyPathEdgesNearEntities']
		else:
			assert isinstance(featureChoice,list)
			for f in featureChoice:
				assert f in validFeatures, "Feature (%s) is not a valid feature" % f
			self.chosenFeatures = featureChoice
		
		self.tfidf = tfidf


		self.dictVectorizers = {}
		self.tfidfTransformers = {}

	def _registerFunctions(self):
		self.featureInfo = {}
		self.featureInfo['entityTypes'] = {'func':_doEntityTypes,'never_tfidf':True}
		self.featureInfo['unigramsBetweenEntities'] = {'func':_doUnigramsBetweenEntities,'never_tfidf':False}
		self.featureInfo['bigrams'] = {'func':_doBigrams,'never_tfidf':False}
		self.featureInfo['dependencyPathEdges'] = {'func':_doDependencyPathEdges,'never_tfidf':True}
		self.featureInfo['dependencyPathEdgesNearEntities'] = {'func':_doDependencyPathEdgesNearEntities,'never_tfidf':True}
				
	def getFeatureNames(self):
		"""
		Get the names for each feature (i.e. each column in matrix generated by the fit_transform() and transform() functions. Fit_transform() must have already been used, i.e. the vectorizer needs to have been fit to training data.
		
		:return: List of names for each feature (column of the vectorized data)
		:rtype: List of str
		"""
		
		assert self.fitted == True, "Must have fit data first"
		featureNames = []
		for feature in self.chosenFeatures:
			if feature in self.dictVectorizers:
				featureNames += self.dictVectorizers[feature].get_feature_names()
		return featureNames
		
	def _vectorize(self,candidates,fit):
		assert isinstance(candidates,list)
		assert len(candidates) > 0
		for c in candidates:
			assert isinstance(c,kindred.CandidateRelation)
			
		matrices = []
		for feature in self.chosenFeatures:
			assert feature in self.featureInfo.keys()
			featureFunction = self.featureInfo[feature]['func']
			never_tfidf = self.featureInfo[feature]['never_tfidf']
			data = featureFunction(candidates,self.entityCount)
			notEmpty = any( len(d)>0 for d in data )
			if fit:
				if notEmpty:
					self.dictVectorizers[feature] = DictVectorizer()
					if self.tfidf and not never_tfidf:
						self.tfidfTransformers[feature] = TfidfTransformer()
						intermediate = self.dictVectorizers[feature].fit_transform(data)
						matrices.append(self.tfidfTransformers[feature].fit_transform(intermediate))
					else:
						matrices.append(self.dictVectorizers[feature].fit_transform(data))
			else:
				if feature in self.dictVectorizers:
					if self.tfidf and not never_tfidf:
						intermediate = self.dictVectorizers[feature].transform(data)
						matrices.append(self.tfidfTransformers[feature].transform(intermediate))
					else:
						matrices.append(self.dictVectorizers[feature].transform(data))

		mergedMatrix = hstack(matrices)
		return mergedMatrix
			
	def fit_transform(self,candidates):
		"""
		Fit the vectorizer to a list of candidate relations found in a corpus and vectorize them to generate the feature matrix.
		
		:param candidates: Relation candidates to vectorize
		:type candidates: list of kindred.CandidateRelation
		:return: Feature matrix (# rows = number of candidate relations, # cols = number of features)
		:rtype: scipy.sparse.csr.csr_matrix
		"""
		assert self.fitted == False
		assert isinstance(candidates,list)
		assert len(candidates) > 0
		for c in candidates:
			assert isinstance(c,kindred.CandidateRelation)
		self.fitted = True
		return self._vectorize(candidates,True)
	
	def transform(self,candidates):
		"""
		Vectorize the candidate relations to generate the feature matrix. Must already have been fit.
		
		:param candidates: Relation candidates to vectorize
		:type candidates: list of kindred.CandidateRelation
		:return: Feature matrix (# rows = number of candidate relations, # cols = number of features)
		:rtype: scipy.sparse.csr.csr_matrix
		"""
		assert self.fitted == True
		assert isinstance(candidates,list)
		assert len(candidates) > 0
		for c in candidates:
			assert isinstance(c,kindred.CandidateRelation)
		return self._vectorize(candidates,False)
		
		
	



import six

class Entity:
	"""
	Biomedical entity with information of location in text

	:ivar entityType: Type of the entity
	:ivar text: Text of the entity
	:ivar position: Position within the text passage at which point entity appears. Entity may be non-contigious
	:ivar sourceEntityID: Entity ID used in source document
	:ivar externalID: ID associated with external ontology (e.g. Hugo Gene ID)
	"""
	
	_nextInternalID = 1

	def __init__(self,entityType,text,position,sourceEntityID=None,externalID=None):
		"""
		Constructor for Entity class
		
		:param entityType: Type of the entity
		:param text: Text of the entity
		:param position: Position within the text passage at which point entity appears. Entity may be non-contigious
		:param sourceEntityID: Entity ID used in source document
		:param externalID: ID associated with external ontology (e.g. Hugo Gene ID)
		:type entityType: str
		:type text: str
		:type position: list of tuples of two integers
		:type sourceEntityID: str
		:type externalID: str
		"""
	
		assert isinstance(entityType, six.string_types), "entityType must be a string"
		assert isinstance(text, six.string_types), "text must be a string"
		assert externalID is None or isinstance(externalID, six.string_types), "externalID must be a string or None"

		posErrorMsg = "Entity position must be list of tuples (startPos,endPos)"
		assert isinstance(position,list), posErrorMsg
		for p in position:
			assert isinstance(p,tuple), posErrorMsg
			assert len(p) == 2, posErrorMsg
			assert isinstance(p[0],int), posErrorMsg
			assert isinstance(p[1],int), posErrorMsg
	
		self.entityType = entityType
		self.sourceEntityID = sourceEntityID
		self.externalID = externalID
		self.text = text
		self.position = position
		
		self.entityID = Entity._nextInternalID
		Entity._nextInternalID += 1
		
	def __str__(self):
		if self.externalID is None:
			out = "<Entity %s:'%s' sourceid=%s %s>" % (self.entityType,self.text,str(self.sourceEntityID),str(self.position))
		else:
			out = "<Entity %s:'%s' sourceid=%s externalid=%s %s>" % (self.entityType,self.text,str(self.sourceEntityID),str(self.externalID),str(self.position))
		return out
		
	def __repr__(self):
		return self.__str__()
		
	def __eq__(self, other):
		"""Override the default Equals behavior"""
		if isinstance(other, self.__class__):
			return self.__dict__ == other.__dict__
		return False
	
	def __ne__(self, other):
		"""Define a non-equality test"""
		return not self.__eq__(other)

	def __hash__(self):
		return hash((self.entityType,self.text,tuple(self.position),self.sourceEntityID,self.externalID))



import kindred
from collections import OrderedDict,defaultdict
import six

# Colors to use for output sentences with annotation
class bcolors:
	HEADER = '\033[95m'
	OKBLUE = '\033[94m'
	OKGREEN = '\033[92m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'

class RESPONSE:
	POSITIVE = 1
	NEGATIVE = 0
	ENTITYERROR = -1
		
	TABLE = {'y':POSITIVE,'n':NEGATIVE,'x':ENTITYERROR}


def manuallyAnnotate(corpus,candidateRelations):
	"""
	Provides a method for basic manual annotation of a series of candidate relations. Deals with a corpus, sentence by sentence, and prompts the user to annotate each candidate relation in turn. Can be exited before completion of the full list and the resulting annotations are split into an annotated corpus and unannotated corpus. Each document in the new corpora are individual sentences.
	
	:param corpus: Corpus of text for annotation
	:param candidateRelations: List of candidate relations (created using CandidateBuilder) to manually review and annotate
	:type corpus: kindred.Corpus
	:type candidateRelations: List of kindred.CandidateRelation
	:return: a tuple of an annotated corpus and unannotated corpus
	:rtype: two kindred.Corpus
	"""

	annotatedCorpus = kindred.Corpus()
	unannotatedCorpus = kindred.Corpus()

	options = OrderedDict()
	options['x'] = 'Done'
	options['0'] = 'None'

	print()
	print("For each sentence, choose an existing option or type the name of a new annotation")
	
	endAnnotation = False
	crCounter = 0
	#for sentence,crsInSentence in groupedBySentences.items():
	for doc in corpus.documents:
		docSentences = set(doc.sentences)
		crsInDoc = [ cr for cr in candidateRelations if cr.sentence in docSentences ]

		doc = kindred.Document(doc.text,doc.entities,[])

		if not endAnnotation:
			for candidateRelation in crsInDoc:
				crCounter += 1

				sentence = candidateRelation.sentence
				sentenceStart = sentence.tokens[0].startPos

				e1,e2 = candidateRelation.entities

				assert len(e1.position) == 1, 'Annotator cannot currently deal with non-continuous entities'
				assert len(e2.position) == 1, 'Annotator cannot currently deal with non-continuous entities'
				start1,end1 = e1.position[0]
				start2,end2 = e2.position[0]

				start1,end1 = start1-sentenceStart,end1-sentenceStart
				start2,end2 = start2-sentenceStart,end2-sentenceStart

				charByChar = list(candidateRelation.sentence.text)
				charByChar[start1] = bcolors.FAIL + charByChar[start1]
				charByChar[end1-1] += bcolors.ENDC
				charByChar[start2] = bcolors.OKGREEN + charByChar[start2]
				charByChar[end2-1] += bcolors.ENDC
				
				sentence = "".join(charByChar)

				print()
				print('#'*30 + " (%d/%d)" % (crCounter,len(candidateRelations)))
				print(sentence)

				optionTxt = " ".join("%s:%s" % (key,value) for key,value in options.items())

				response = None
				while not response:
					response = six.moves.input('%s ? ' % optionTxt).strip()

				if response == 'x':
					endAnnotation = True
					break
				elif response and not response in optionTxt:
					newKey = str(len(options)-1)
					options[newKey] = response
				else:
					response = options[response]

				if response != 'None':
					r = kindred.Relation(response,candidateRelation.entities)
					doc.addRelation(r)

		if endAnnotation:
			# Annotation is incomplete, so wipe any previous annotation on this sentence
			doc.relations = []
			unannotatedCorpus.addDocument(doc)
		else:
			annotatedCorpus.addDocument(doc)

	return annotatedCorpus,unannotatedCorpus





"""
Importer for PubTator data
"""

import kindred
import requests
import json
import time

def _loadPMID(pmid,retries=3):
	assert isinstance(pmid,int)
	
	annotationsURL = "https://www.ncbi.nlm.nih.gov/CBBresearch/Lu/Demo/RESTful/tmTool.cgi/BioConcept/%d/JSON" % pmid

	success = False
	for retry in range(retries):
		try:
			request = requests.get(annotationsURL)
			annotations = request.json()
			success = True
			break
		except ValueError as e:
		 	time.sleep(1)

	if not success:
		raise RuntimeError('Unable to download PubTator data after %d retries' % retries)

	doc = kindred.loadFunctions.parseJSON(annotations)
	
	return doc

def load(pmids):
	"""
	Load a set of documents with annotations from Pubmed given a list of Pubmed IDs (PMIDs)
	
	>>> corpus = load(19894120)
	>>> len(corpus.documents)
	1

	:param pmids: the list of Pubmed IDs
	:type pmids: List of ints
	:returns: a kindred corpus object
	:rtype: kindred.Corpus
	"""

	assert isinstance(pmids,list) or isinstance(pmids,int)

	corpus = kindred.Corpus()
	if isinstance(pmids,list):
		for pmid in pmids:
			doc = _loadPMID(pmid)
			assert isinstance(doc,kindred.Document)
			corpus.addDocument(doc)
	elif isinstance(pmids,int):
		doc = _loadPMID(pmids)
		assert isinstance(doc,kindred.Document)
		corpus.addDocument(doc)
	return corpus
	
	
	
	




class Token:
	"""
	Individual word with lemma, part-of-speech and location in text.
	
	:ivar word: Unprocessed word
	:ivar lemma: Lemmatized word
	:ivar partofspeech: Part-of-speech of word
	:ivar startPos: Start position of token in document text (note: not the sentence text)
	:ivar endPos: End position of token in document text (note: not the sentence text)
	"""
	
	def __init__(self,word,lemma,partofspeech,startPos,endPos):
		"""
		Constructor for Token class
		
		:param word: Unprocessed word
		:param lemma: Lemmatized word
		:param partofspeech: Part-of-speech of word
		:param startPos: Start position of token in document text (note: not the sentence text)
		:param endPos: End position of token in document text (note: not the sentence text)
		:type word: str
		:type lemma: str
		:type partofspeech: str
		:type startPos: int
		:type endPos: int
		"""

		self.word = word
		self.lemma = lemma
		self.partofspeech = partofspeech
		self.startPos = startPos
		self.endPos = endPos

	def __str__(self):
		return self.word
		
	def __repr__(self):
		return self.__str__()




import kindred
import six

class CandidateRelation:
	"""
	Describes a candidate relation between entities (i.e. one that could exist but has not yet been predicted). Contains information about known relation types and arg names associated with this candidate (from training data) and also a link to the sentence containing this candidate.
	
	:ivar entities: List of entities in relation
	:ivar knownTypesAndArgNames: List of tuples with known relation types and argument names associated with this candidate relation
	:ivar sentence: Parsed sentence containing the candidate relation
	"""
	
	def __init__(self,entities=[],knownTypesAndArgNames=[],sentence=None):
		"""
		Constructor for Candidate Relation class
		
		:param entities: List of entities in relation
		:param knownTypesAndArgNames: List of tuples with known relation types and argument names associated with this candidate relation
		:param sentence: Parsed sentence containing the candidate relation
		:type entities: list of kindred.Entity
		:type knownTypesAndArgNames: list of tuples (str, list of str)
		:type sentence: kindred.Sentence
		"""

		assert isinstance(entities,list),  "entities must be a list of kindred.Entity"
		for entity in entities:
			assert isinstance(entity, kindred.Entity), "entities must be a list of kindred.Entity"
		self.entities = entities

		knownTypesAndArgNamesError = "knownTypesAndArgNames must be a list of tuples where each (length=2) tuple is the name of the relation and a list of argument names"
		assert isinstance(knownTypesAndArgNames,list), knownTypesAndArgNamesError
		for knownTypeAndArgNames in knownTypesAndArgNames:
			assert isinstance(knownTypeAndArgNames, tuple), knownTypesAndArgNamesError
			assert len(knownTypeAndArgNames) == 2, knownTypesAndArgNamesError
			knownType,knownArgNames = knownTypeAndArgNames
			assert isinstance(knownType, six.string_types), knownTypesAndArgNamesError
			assert isinstance(knownArgNames,list), knownTypesAndArgNamesError
			for knownArgName in knownArgNames:
				assert isinstance(knownArgName, six.string_types), knownTypesAndArgNamesError
		self.knownTypesAndArgNames = knownTypesAndArgNames
				
		assert sentence is None or isinstance(sentence, kindred.Sentence)
		self.sentence = sentence
	
	def __eq__(self, other):
		"""Override the default Equals behavior"""
		if isinstance(other, self.__class__):
			return self.__dict__ == other.__dict__
		return False
	
	def __ne__(self, other):
		"""Define a non-equality test"""
		return not self.__eq__(other)

	def __str__(self):
		return "<CandidateRelation %s %s>" % (str(self.entities),str(self.knownTypesAndArgNames))

	def __repr__(self):
		return self.__str__()

	def __hash__(self):
		knownTypesAndArgNamesToTuple = tuple([ (knownType,tuple(knownArgNames)) for knownType,knownArgNames in self.knownTypesAndArgNames ])

		return hash((tuple(self.entities),knownTypesAndArgNamesToTuple,self.sentence))




import kindred
import networkx as nx
import itertools
import sys
import six
from collections import defaultdict

class Sentence:
	"""
	Set of tokens for a sentence after parsing
	
	:ivar text: Text of the sentence
	:ivar tokens: List of tokens in sentence
	:ivar dependencies: List of dependencies from dependency path. Should be a list of tuples with form (tokenindex1,tokenindex2,dependency_type)
	:ivar sourceFilename: Filename of the source document
	:ivar entityAnnotations: List of entities associated with token indices
	"""
	
	def __init__(self, text, tokens, dependencies, sourceFilename=None):
		"""
		Constructor for Sentence class
	
		:param text: Text of the sentence
		:param tokens: List of tokens in sentence
		:param dependencies: List of dependencies from dependency path. Should be a list of tuples with form (tokenindex1,tokenindex2,dependency_type)
		:param sourceFilename: Filename of the source document
		:type text: str
		:type tokens: list of kindred.Token
		:type dependencies: list of tuples
		:type sourceFilename: str
		"""

		assert isinstance(text, six.string_types)

		assert isinstance(tokens, list)
		for token in tokens:
			assert isinstance(token,kindred.Token)
		
		# Check the format of the Dependencies
		dependencyErrorMsg = "Each dependency is expected to be a tuple of (tokenindex1,tokenindex2,dependency_type). Token index can be -1 to indicate an incoming edge."
		assert isinstance(dependencies, list), dependencyErrorMsg
		for dependency in dependencies:
			assert isinstance(dependency,tuple),dependencyErrorMsg
			assert len(dependency) == 3,dependencyErrorMsg
			assert isinstance(dependency[0],int),dependencyErrorMsg
			assert isinstance(dependency[1],int),dependencyErrorMsg
			assert isinstance(dependency[2], six.string_types),dependencyErrorMsg
			assert dependency[0] >= -1 and dependency[0] < len(tokens), dependencyErrorMsg
			assert dependency[1] >= -1 and dependency[1] < len(tokens), dependencyErrorMsg
		
		self.text = text
		self.tokens = tokens
		self.dependencies = dependencies
		self.sourceFilename = sourceFilename
		
		self.entityAnnotations = []
		
	def addEntityAnnotation(self, entity, tokenIndices):
		"""
		Add an entity annotation to this sentence. Associated a specific entity with the indices of specific tokens
		
		:param entity: Entity to add to sentence
		:param tokenIndices: List of token indices
		:type entity: kindred.Entity
		:type tokenIndices: List of ints
		"""
		
		assert isinstance(entity, kindred.Entity)
		assert isinstance(tokenIndices,list)
		for l in tokenIndices:
			assert l >= 0 and l < len(self.tokens), "Entity location must be an index of one of the tokens"
		self.entityAnnotations.append( (entity,tokenIndices) )
	
	def __str__(self):
		tokenWords = [ t.word for t in self.tokens ]
		return " ".join(tokenWords)
	
	def __repr__(self):
		return self.__str__()
	
	def extractMinSubgraphContainingNodes(self, minSet):
		"""
		Find the minimum subgraph of the dependency graph that contains the provided set of nodes. Useful for finding dependency-path like structures
		
		:param minSet: List of token indices
		:type minSet: List of ints
		:return: All the nodes and edges in the minimal subgraph
		:rtype: Tuple of nodes,edges where nodes is a list of token indices, and edges are the associated dependency edges between those tokens
		"""

		assert isinstance(minSet, list)
		for i in minSet:
			assert isinstance(i, int)
			assert i >= 0
			assert i < len(self.tokens)
		G1 = nx.Graph()
		for a,b,depType in self.dependencies:
			G1.add_edge(a,b,dependencyType=depType)

		G2 = nx.Graph()
		paths = {}

		minSet = sorted(list(set(minSet)))
		setCount1 = len(minSet)
		minSet = [ a for a in minSet if G1.has_node(a) ]
		setCount2 = len(minSet)
		if setCount1 != setCount2:
			sys.stderr.write("WARNING. %d node(s) not found in dependency graph!\n" % (setCount1-setCount2))
		for a,b in itertools.combinations(minSet,2):
			try:
				path = nx.shortest_path(G1,a,b)
				paths[(a,b)] = path
				G2.add_edge(a,b,weight=len(path))
			except nx.exception.NetworkXNoPath:
				sys.stderr.write("WARNING. No path found between nodes %d and %d!\n" % (a,b))
			
		# TODO: This may through an error if G2 ends up having multiple components. Catch it gracefully.
		minTree = nx.minimum_spanning_tree(G2)
		nodes = set()
		allEdges = set()
		for a,b in minTree.edges():
			path = paths[(min(a,b),max(a,b))]
			for i in range(len(path)-1):
				a,b = path[i],path[i+1]
				dependencyType = G1.get_edge_data(a,b)['dependencyType']
				edge = (min(a,b),max(a,b),dependencyType)
				allEdges.add(edge)
			nodes.update(path)

		return nodes,allEdges





import numpy as np
from sklearn.linear_model import LogisticRegression

class LogisticRegressionWithThreshold:
	"""
	A modified Logistic Regression classifier that will filter calls by a custom threshold, instead of the default 0.5. This allows for control of the precision-recall tradeoff, e.g. false positives versus false negatives.
	
	:ivar clf: The underlying LogisticRegression classifier
	:ivar threshold: Threshold to use, should be between 0 and 1
	"""

	def __init__(self,threshold=0.5):
		"""
		Set up a Logistic Regression classifier that can use a different threshold for predictions and thereby be more lenient (lower threshold, false positives increase, false negatives decrease) or more conservative (higher threshold, false positives decrease, false negative increase).
		
		:param threshold: Threshold to use, should be between 0 and 1
		:type threshold: float
		"""
		
		assert threshold >= 0 and threshold <= 1, "Threshold must be between 0 and 1"

		self.clf = LogisticRegression(class_weight='balanced',random_state=1,solver='liblinear',multi_class='ovr')
		self.threshold = threshold

	def fit(self,X,Y):
		"""
		Train the classifier using the associated matrix X and classes Y. Class zero should represent no associated class.
		
		:param X: Training vector
		:param Y: Associated class for each row of X
		:type X: sparse matrix
		:type Y: matrix
		"""

		self.clf.fit(X,Y)
		self.classes_ = self.clf.classes_

	def predict(self,X):
		"""
		Make predictions for the class of each row in X. Class zero should represent no prediction.
		
		:param X: Testing vector
		:type X: sparse matrix
		:return: Predictions of classes for each row in X
		:rtype: matrix
		"""

		probs = self.clf.predict_proba(X)

		# Ignore probabilities that fall below our threshold
		probs[probs<self.threshold] = -1.0

		# Make sure that the zero class is only select if all other options are below the threshold
		probs[:,0] = -0.5

		# And get the highest probability for each row
		predictions = np.argmax(probs,axis=1)

		return predictions

	def predict_proba(self,X):
		"""
		Calculate probabilities for the class of each row in X. Class zero should represent no prediction.
		Returns a matrix of probabilities
		
		:param X: Testing vector
		:type X: sparse matrix
		:return: Probabilities of classes for each row in X
		:rtype: matrix
		"""

		return self.clf.predict_proba(X)





import os
import zipfile
import hashlib
import requests
import logging
import traceback

def _calcSHA256(filename):
	return hashlib.sha256(open(filename, 'rb').read()).hexdigest()

def _findDir(name, path):
	if os.path.isdir(path):
		for root, dirs, files in os.walk(path):
			if name in dirs:
				return os.path.abspath(os.path.join(root, name))
	return None

# From: https://stackoverflow.com/questions/32763720/timeout-a-file-download-with-python-urllib
def _downloadFile(url,filename,timeout=180):
	# Make the actual request, set the timeout for no data to 10 seconds and enable streaming responses so we don't have to keep the large files in memory
	request = requests.get(url, timeout=10, stream=True)

	# Open the output file and make sure we write in binary mode
	with open(filename, 'wb') as fh:
		# Walk through the request response in chunks of 1024 * 1024 bytes, so 1MiB
		for chunk in request.iter_content(1024 * 1024):
			# Write the chunk to the file
			fh.write(chunk)
			

def _downloadFiles(files,downloadDirectory):
	if not os.path.isdir(downloadDirectory):
		os.mkdir(downloadDirectory)

	for url,shortName,expectedSHA256 in files:
		downloadedPath = os.path.join(downloadDirectory,shortName)
	
		if os.path.isfile(downloadedPath):
			downloadedSHA256 = _calcSHA256(downloadedPath)
			if not downloadedSHA256 == expectedSHA256:
				os.remove(downloadedPath)

		if not os.path.isfile(downloadedPath):
			try:
				_downloadFile(url,downloadedPath)
			except Exception as e:
				logging.error(traceback.format_exc())
				print(type(e))
				raise
			
			downloadedSHA256 = _calcSHA256(downloadedPath)
			assert downloadedSHA256 == expectedSHA256, "SHA256 mismatch with downloaded file: %s" % shortName
			
			if shortName.endswith('.zip'):
				zip_ref = zipfile.ZipFile(downloadedPath, 'r')
				zip_ref.extractall(path=downloadDirectory)
				zip_ref.close()

				os.remove(downloadedPath)
	



import sys
import itertools
import kindred
import pickle
import argparse
import codecs
import time
import re
import string
from collections import defaultdict,Counter
import json
import six
import os

def acronymMatch(words,pos,currentAcronym,atStart,subpos=None):
	if len(currentAcronym) == 0:
		if not (subpos is None): # Can't finish acronym mid-word
			return []
		else:
			return [pos+1]

	curWord = words[pos].lower()
	wordSplit = curWord.split('-')
	curLetter = currentAcronym[-1]
	
	moves = []
	
	if subpos is None:
		if atStart and curLetter == 's' and curWord[-1] == 's':
			# Possible plural
			moves.append( (words,pos,currentAcronym[:-1],False) )
			
		if curLetter == curWord[0]:
			moves.append( (words,pos-1,currentAcronym[:-1],False) )

		if curWord == '-':
			moves.append( (words,pos-1,currentAcronym,False) )
			

	if len(wordSplit) > 1:
		if subpos is None:
			subpos = len(wordSplit)-1
		if len(wordSplit[subpos]) > 0 and curLetter == wordSplit[subpos][0]:
			if subpos == 0:
				moves.append( (words,pos-1,currentAcronym[:-1],False) )
			else:
				moves.append( (words,pos,currentAcronym[:-1],False,subpos-1) )
			
	possibleStarts = []
	for move in moves:
		possibleStarts += acronymMatch(*move)
		
	return possibleStarts

def acronymDetection(words):
	LRBs = [i for i, x in enumerate(words) if x == u'(']
	RRBs = [i for i, x in enumerate(words) if x == u')']
	acronyms = []
	for i,j in itertools.product(LRBs,RRBs):
		if j-i == 2:
			acronymLoc = i+1
			possibleAcronym = words[acronymLoc]
			possibleStarts = acronymMatch(words,i-1,possibleAcronym.lower(),True)
			if len(possibleStarts) > 0:
				start = min(possibleStarts)
				end = i
				acronyms.append((start,end,acronymLoc))
	return acronyms

def mergeWordsForFusionDetection(words):
	prevWord = ""
	mergedWords = []
	start = 0
	mergeChars = ['-','/',':']
	for i,w in enumerate(words):
		if w in mergeChars:
			prevWord += w
		elif len(prevWord) > 0 and prevWord[-1] in mergeChars:
			prevWord += w
		else:
			if prevWord:
				mergedWords.append((start,i-1,prevWord))
			prevWord = w
			start = i

	if prevWord:
		mergedWords.append((start,len(words)-1,prevWord))

	return mergedWords

def fusionGeneDetection(words, lookupDict):
	termtypesAndids,terms,locs = [],[],[]
	origWords = list(words)
	words = [ w.lower() for w in words ]

	mergedWords = mergeWordsForFusionDetection(words)

	for start,end,word in mergedWords:
		split = re.split("[-/:]",word)
		fusionCount = len(split)
		if fusionCount == 1:
			continue
			
		allGenes = True
		
		geneIDs = []
		lookupIDCounter = Counter()
		for s in split:
			key = s
			if key in lookupDict:
				isGene = False
				for entityType,entityID in lookupDict[key]:
					if entityType == 'gene':
						for tmpID in entityID.split(';'):
							lookupIDCounter[tmpID] += 1

						geneIDs.append(entityID)
						isGene = True
						break
				if not isGene:
					allGenes = False
					break
			else:
				allGenes = False
				break

		# We're going to check if there are any lookup IDs shared among all the "fusion" terms
		# Hence this may not actually be a fusion, but just using multiple names of a gene
		# e.g. HER2/neu
		completeLookupIDs = [ id for id,count in lookupIDCounter.items() if count == fusionCount ]
		if len(completeLookupIDs) > 0:
			termtypesAndids.append([('gene',';'.join(completeLookupIDs))])
			terms.append(tuple(origWords[start:end+1]))
			locs.append((start,end+1))
		elif allGenes: # All the terms look like genes (and different genes), so we're going to mark this as a fusion (or combo)
			#geneTxt = ",".join(map(str,geneIDs))
			geneIDs = [ geneID.replace(';','&') for geneID in geneIDs ]
			termtypesAndids.append([('gene','combo|' + '|'.join(geneIDs))])
			terms.append(tuple(origWords[start:end+1]))
			locs.append((start,end+1))
		
	return locs,terms,termtypesAndids

def getTermIDsAndLocations(sentence, lookupDict):
	termtypesAndids,terms,locs = [],[],[]
	# Lowercase all the tokens
	#np = [ unicodeLower(w) for w in np ]
	#orignp = np
	np = [ t.word.lower() for t in sentence.tokens ]
	blank = "".join( " " for _ in sentence.text )

	tempSentence = sentence.text.lower()
	sentenceStart = sentence.tokens[0].startPos
	# The length of each search string will decrease from the full length
	# of the text down to 1
	for l in reversed(range(1, len(sentence.tokens)+1)):
		# We move the search window through the text
		for i in range(len(np)-l+1):
			# Extract that window of text
			#s = tuple(np[i:i+l])
			startPos = sentence.tokens[i].startPos - sentenceStart
			endPos = sentence.tokens[i+l-1].endPos - sentenceStart
			s = tempSentence[startPos:endPos]

			# Search for it in the dictionary
			if s in lookupDict:
				# If found, save the ID(s) in the dictionary
				termtypesAndids.append(lookupDict[s])
				terms.append(tuple(np[i:i+l]))
				locs.append((i,i+l))
				# And blank it out
				#np[i:i+l] = [ "" for _ in range(l) ]
				tempSentence = tempSentence[:startPos] + blank[startPos:endPos] + tempSentence[endPos:]

	# Then return the found term IDs
	return locs,terms,termtypesAndids

def startsWithButNotAll(s,search):
	return s.startswith(search) and len(s) > len(search)

def cleanupVariant(variant):
	variant = variant.upper().replace('P.','')
	aminoAcidInfo = [('ALA','A'),('ARG','R'),('ASN','N'),('ASP','D'),('CYS','C'),('GLU','E'),('GLN','Q'),('GLY','G'),('HIS','H'),('ILE','I'),('LEU','L'),('LYS','K'),('MET','M'),('PHE','F'),('PRO','P'),('SER','S'),('THR','T'),('TRP','W'),('TYR','Y'),('VAL','V')]
	for longA,shortA in aminoAcidInfo:
		variant = variant.replace(longA,shortA)
	return variant

class EntityRecognizer:
	"""
	Annotates entities in a Corpus using an exact-dictionary matching scheme with additional heuristics. These heuristics include detecthing fusion gene mentions, microRNA, identifying acronyms to reduce ambiguity, identifying variants and more. All the options are parameters for the constructor of this class.
	
	:ivar lookup: Used for the dictionary matching. A dictionary of terms (tuple of parsed words) to a list of (entityType,externalID).
	:ivar detectFusionGenes: Whether it will try to identify fusion gene terms (e.g. BCR-ABL1). Lookup must contain terms of type 'gene'
	:ivar detectMicroRNA: Whether it will identify microRNA terms (added as 'gene' entities)
	:ivar acronymDetectionForAmbiguity: Whether it will try to identify acronyms and use this to deal with ambiguity (by removing incorrect matches to acronyms or the longer terms)
	:ivar mergeTerms: Whether it will merge neighbouring terms that refer to the same external entity (e.g. HER2/neu as one term instead of two)
	:ivar detectVariants: Whether it will identify a variant (e.g. V600E) and create an entity of type 'variant'
	:ivar variantStopwords: Variant terms to be ignored (e.g. S100P) if detectVariants is used
	:ivar detectPolymorphisms: Whether it will identify a SNP (using a dbSNP ID) and create an entity of type 'variant'
	:ivar removePathways: Whether it will remove genes that are actually naming a signalling pathway (e.g. MTOR pathway)
	"""

	def __init__(self,lookup,detectFusionGenes=False,detectMicroRNA=False,acronymDetectionForAmbiguity=False,mergeTerms=False,detectVariants=False,variantStopwords=[],detectPolymorphisms=False,removePathways=False):
		"""
		Create an EntityRecognizer and provide the lookup table for terms and additional flags for what to identify in text

		:param lookup: A dictionary of terms (tuple of parsed words) to a list of (entityType,externalID).
		:param detectFusionGenes: Whether to try to identify fusion gene terms (e.g. BCR-ABL1). Lookup must contain terms of type 'gene'
		:param detectMicroRNA: Whether to identify microRNA terms (added as 'gene' entities)
		:param acronymDetectionForAmbiguity: Whether to try to identify acronyms and use this to deal with ambiguity (by removing incorrect matches to acronyms or the longer terms)
		:param mergeTerms: Whether to merge neighbouring terms that refer to the same external entity (e.g. HER2/neu as one term instead of two)
		:param detectVariants: Whether to identify a variant (e.g. V600E) and create an entity of type 'variant'
		:param variantStopwords: Variant terms to be ignored (e.g. S100P) if detectVariants is used
		:param detectPolymorphisms: Whether to identify a SNP (using a dbSNP ID) and create an entity of type 'variant'
		:param removePathways: Remove genes that are actually naming a signalling pathway (e.g. MTOR pathway)
		:type lookup: dict
		:type detectFusionGenes: bool
		:type detectMicroRNA: bool
		:type acronymDetectionForAmbiguity: bool
		:type mergeTerms: bool
		:type detectVariants: bool
		:type variantStopwords: list
		:type detectPolymorphisms: bool
		:type removePathways: bool
		"""

		assert isinstance(lookup,dict)
		for termsmatch,typeAndIDs in lookup.items():
			assert isinstance(termsmatch,six.string_types), "Lookup key must be a tuple of strings"
			assert isinstance(typeAndIDs,set), "Lookup value must be a list of (entityType,externalID)"
			assert len(typeAndIDs)>0, "Lookup value must be a list of (entityType,externalID)"
			for typeAndID in typeAndIDs:
				assert isinstance(typeAndID,tuple),"Lookup value must be a list of (entityType,externalID)"
				assert len(typeAndID)==2, "Lookup value must be a list of (entityType,externalID)"

		assert isinstance(detectFusionGenes,bool)
		assert isinstance(detectMicroRNA,bool)
		assert isinstance(acronymDetectionForAmbiguity,bool)
		assert isinstance(mergeTerms,bool)
		assert isinstance(detectVariants,bool)
		assert isinstance(detectPolymorphisms,bool)
		
		assert isinstance(variantStopwords,list)
		for variantStopword in variantStopwords:
			assert isinstance(variantStopword,six.string_types), "variantStopwords should be a list of strings"

		self.lookup = lookup
		self.detectFusionGenes = detectFusionGenes
		self.detectMicroRNA = detectMicroRNA
		self.acronymDetectionForAmbiguity = acronymDetectionForAmbiguity
		self.mergeTerms = mergeTerms
		self.detectVariants = detectVariants
		self.variantStopwords = set(variantStopwords)
		self.detectPolymorphisms = detectPolymorphisms
		self.removePathways = removePathways

		
	def _processWords(self, sentence):
		locs,terms,termtypesAndids = getTermIDsAndLocations(sentence,self.lookup)

		words = [ t.word for t in sentence.tokens ]

		if self.detectVariants:
			#snvRegex = r'^[A-Z][0-9]+[A-Z]$'
			variantRegex1 = r'^[ACDEFGHIKLMNPQRSTVWY][1-9][0-9]*[ACDEFGHIKLMNPQRSTVWY]$'
			variantRegex2 = r'^(p\.)?((Ala)|(Arg)|(Asn)|(Asp)|(Cys)|(Glu)|(Gln)|(Gly)|(His)|(Ile)|(Leu)|(Lys)|(Met)|(Phe)|(Pro)|(Ser)|(Thr)|(Trp)|(Tyr)|(Val))[1-9][0-9]*((Ala)|(Arg)|(Asn)|(Asp)|(Cys)|(Glu)|(Gln)|(Gly)|(His)|(Ile)|(Leu)|(Lys)|(Met)|(Phe)|(Pro)|(Ser)|(Thr)|(Trp)|(Tyr)|(Val))$'

			filteredWords = [ w for w in words if not w in self.variantStopwords ]
			snvMatches1 = [ not (re.match(variantRegex1,w) is None) for w in filteredWords ]
			snvMatches2 = [ not (re.match(variantRegex2,w,re.IGNORECASE) is None) for w in filteredWords ]

			snvMatches = [ (match1 or match2) for match1,match2 in zip(snvMatches1,snvMatches2) ]
			for i,(w,snvMatch) in enumerate(zip(words,snvMatches)):
				if snvMatch:
					cleaned = cleanupVariant(w)
					potentialLocs = (i,i+1)
					if not potentialLocs in locs:
						termtypesAndids.append([('variant',"substitution|%s"%cleaned)])
						terms.append((w,))
						locs.append(potentialLocs)

		if self.detectPolymorphisms:
			polymorphismRegex1 = r'^rs[1-9][0-9]*$'

			polyMatches = [ not (re.match(polymorphismRegex1,w) is None) for w in words ]

			for i,(w,polyMatch) in enumerate(zip(words,polyMatches)):
				if polyMatch:
					potentialLocs = (i,i+1)
					if not potentialLocs in locs:
						termtypesAndids.append([('variant','dbsnp|%s'%w)])
						terms.append((w,))
						locs.append(potentialLocs)

		if self.detectMicroRNA:
			for i,w in enumerate(words):
				lw = w.lower()
				if startsWithButNotAll(lw,"mir-") or startsWithButNotAll(lw,"hsa-mir-") or startsWithButNotAll(lw,"microrna-") or (startsWithButNotAll(lw,"mir") and lw[3] in string.digits):
					potentialLocs = (i,i+1)
					if not potentialLocs in locs:
						termtypesAndids.append([('gene','mirna|'+w)])
						terms.append((w,))
						locs.append((i,i+1))

		toRemove = []
		if self.detectFusionGenes:
			fusionLocs,fusionTerms,fusionTermtypesAndids = fusionGeneDetection(words,self.lookup)
	
			for floc,fterm,ftermtypesAndid in zip(fusionLocs,fusionTerms,fusionTermtypesAndids):
				if not floc in locs:
					# Check for which entities to remove that are inside this fusion term
					fstart,fend = floc
					for tstart,tend in locs:
						if fstart <= tstart and tend <= fend:
							toRemove.append((tstart,tend))

					locs.append(floc)
					terms.append(fterm)
					termtypesAndids.append(ftermtypesAndid)

		filtered = zip(locs,terms,termtypesAndids)
		filtered = [ (l,t,ti) for l,t,ti in filtered if not l in toRemove ]
		filtered = sorted(filtered)

		if self.mergeTerms:
			# We'll attempt to merge terms (i.e. if a gene is referred to using two acronyms together)
			# Example: Hepatocellular carcinoma (HCC) or HER2/ Neu or INK4B P15
			locsToRemove = set()
			for i in range(len(filtered)-1):
				(startA,endA),termsA,termTypesAndIDsA = filtered[i]
				(startB,endB),termsB,termTypesAndIDsB = filtered[i+1]
				
				# Check that the terms are beside each other or separated by a /,- or (
				if startB == endA or (startB == (endA+1) and words[endA] in ['/','-','(',')']):
					idsA,idsB = set(),set()

					for termType, termIDs in termTypesAndIDsA:
						for termID in termIDs.split(';'):
							idsA.add((termType,termID))
					for termType, termIDs in termTypesAndIDsB:
						for termID in termIDs.split(';'):
							idsB.add((termType,termID))

					idsIntersection = idsA.intersection(idsB)

					# Detect if the either term is in brackets e.g. HER2 (ERBB2)
					firstTermInBrackets,secondTermInBrackets = False,False
					if startB == (endA+1) and endB < len(words) and words[endA] == '(' and words[endB] == ')':
						secondTermInBrackets = True
					if startB == (endA+1) and startA > 0 and words[startA-1] == '(' and words[endA] == ')':
						firstTermInBrackets = True

					# The two terms share IDs so we're going to merge them
					idsShared = (len(idsIntersection) > 0)

					if idsShared:
						groupedByType = defaultdict(list)
						for termType,termID in idsIntersection:
							groupedByType[termType].append(termID)

						locsToRemove.add((startA,endA))
						locsToRemove.add((startB,endB))

						if secondTermInBrackets:
							thisLocs = (startA,endB+1)
							thisTerms = tuple(words[startA:endB+1])
						elif firstTermInBrackets:
							thisLocs = (startA-1,endB)
							thisTerms = tuple(words[startA-1:endB])
						else:
							thisLocs = (startA,endB)
							thisTerms = tuple(words[startA:endB])


						thisTermTypesAndIDs = [ (termType,";".join(sorted(termIDs))) for termType,termIDs in groupedByType.items() ]

						filtered.append((thisLocs,thisTerms,thisTermTypesAndIDs))

			# Now we have to remove the terms marked for deletion in the previous section
			filtered = [ (locs,terms,termtypesAndids) for locs,terms,termtypesAndids in filtered if not locs in locsToRemove]
			filtered = sorted(filtered)

		if self.acronymDetectionForAmbiguity:
			# And we'll check to see if there are any obvious acronyms
			locsToRemove = set()
			acronyms = acronymDetection(words)
			for (wordsStart,wordsEnd,acronymLoc) in acronyms:
				wordIsTerm = (wordsStart,wordsEnd) in locs
				acronymIsTerm = (acronymLoc,acronymLoc+1) in locs
				
				if wordIsTerm and acronymIsTerm:
					# Remove the acronym
					locsToRemove.add((acronymLoc,acronymLoc+1))
				elif acronymIsTerm:
					# Remove any terms that contain part of the spelt out word
					newLocsToRemove = [ (i,j) for i in range(wordsStart,wordsEnd) for j in range(i,wordsEnd+1) ]
					locsToRemove.update(newLocsToRemove)
				

			# Now we have to remove the terms marked for deletion in the previous section
			filtered = [ (locs,terms,termtypesAndids) for locs,terms,termtypesAndids in filtered if not locs in locsToRemove]
			filtered = sorted(filtered)

		if self.removePathways:
			forbiddenPathwayWords = set(['pathway','pathways','signaling','signalling','cascade'])
			filtered2 = []
			for locs,terms,termtypesAndids in filtered:
				nextTokenIndex = locs[1]
				nextTokenIsForbiddenWord = nextTokenIndex < len(words) and words[nextTokenIndex].lower() in forbiddenPathwayWords
				if nextTokenIsForbiddenWord:
					termtypesAndids = [ (termtype,termid) for termtype,termid in termtypesAndids if not termtype == 'gene' ]
				if len(termtypesAndids) > 0:
					filtered2.append((locs,terms,termtypesAndids))

			filtered = filtered2
		
		return filtered

	def annotate(self,corpus):
		"""
		Annotate a parsed corpus with the wordlist lookup and other entity types

		:param corpus: Corpus to annotate
		:type corpus: kindred.Corpus
		"""

		assert corpus.parsed == True, "Corpus must already be parsed before entity recognition"

		for doc in corpus.documents:
			entityCount = 0
			for sentence in doc.sentences:

				extractedTermData = self._processWords(sentence)
				
				for locs,terms,termtypesAndids in extractedTermData:
					startToken = locs[0]
					endToken = locs[1]
					startPos = sentence.tokens[startToken].startPos
					endPos = sentence.tokens[endToken-1].endPos
					text = doc.text[startPos:endPos]
					loc = list(range(startToken,endToken))
					for entityType,externalID in termtypesAndids:
						sourceEntityID = "T%d" % (entityCount+1)

						e = kindred.Entity(entityType,text,[(startPos,endPos)],externalID=externalID,sourceEntityID=sourceEntityID)
						doc.addEntity(e)
						sentence.addEntityAnnotation(e,loc)
						entityCount += 1

	@staticmethod
	def loadWordlists(entityTypesWithFilenames, idColumn=0, termsColumn=1, columnSeparator='\t', termSeparator='|'):
		"""
		Load a wordlist from multiple files. By default, each file should be a tab-delimited file with the first column is the ID and the second column containing all the terms separated by '|'. This can be modified by the parameters.

		As each term is parsed, this can take a long time. It is recommended to run this one time and save the output as a Python pickle file and load in.

		:param entityTypesWithFilenames: Dictionary of entityType => filename
		:param idColumn: The column containing the ID for the term (starts from 0)
		:param termsColumn: The column containing the list of terms (starts from 0)
		:param columnSeparator: The column separator for the file (default is a tab)
		:param termSeparator: The separator for the list of terms (default is a '|')
		:type entityTypesWithFilenames: dict
		:type idColumn: int
		:type termsColumn: int
		:type columnSeparator: str
		:type termSeparator: str
		:return: Dictionary of lookup values
		:rtype: dict
		"""
		errorMsg = 'entityTypesWithFilenames should be a dictionary with pairs of {entityType: filename} where both are strings and the filename points to a file that exists'
		assert isinstance(entityTypesWithFilenames,dict), errorMsg
		for entityType,filename in entityTypesWithFilenames.items():
			assert isinstance(entityType,six.string_types), errorMsg
			assert isinstance(filename,six.string_types), errorMsg
			assert os.path.isfile(filename), "%s does not exist" % filename

		assert isinstance(idColumn,int)
		assert isinstance(termsColumn,int)
		assert isinstance(columnSeparator,str)
		assert isinstance(termSeparator,str)

		requiredColumns = max(idColumn,termsColumn)+1

		lookup = defaultdict(set)
		for entityType,filename in entityTypesWithFilenames.items():
			with codecs.open(filename,'r','utf-8') as f:
				tempLookup = defaultdict(set)
				for lineno,line in enumerate(f):
					split = line.strip().split(columnSeparator)
					assert len(split) >= requiredColumns, 'Line %d contains only %d columns when %d are required' % (lineno+1,len(split),requiredColumns)
					termid,terms = split[idColumn],split[termsColumn]
					for term in terms.split(termSeparator):
						tempLookup[term.strip()].add(termid)

			for tupleterm,idlist in tempLookup.items():
				lookup[tupleterm].add( (entityType,";".join(sorted(list(idlist)))) )

		return dict(lookup)




import sklearn
import numpy as np
import inspect

class MultiLabelClassifier:
	"""
	Wrapper for a set of classifiers that can behave as a multi-label classifier. Multi-label means that each data point can have multiple labels (or belong to multiple classes). This is particularly relevant in text mining where two words can belong to multiple relations. This class just creates a classifier for each label and runs then together, concatenating the results into a nice matrix form
	"""
	def __init__(self,classifier,**kwargs):
		"""
		Create a classifier that can handle multiple labels using multiple instance of the supplied classifier class. Any additional parameters are passed onto the classifier.

		:param classifier: The type of classifier to use
		:type classifier: class with fit/predict
		"""

		assert inspect.isclass(classifier), "Must provide a class with fit/predict methods"

		assert hasattr(classifier, 'fit') and callable(getattr(classifier, 'fit')), "Provided classifier class must have a fit method"
		assert hasattr(classifier, 'predict') and callable(getattr(classifier, 'predict')), "Provided classifier class must have a predict method"

		self.classifierClass = classifier
		self.classifierArgs = kwargs

		self.classifiers = None
		self.labelCount = None
		self.fitted = False

	def fit(self,X,Y):
		"""
		Fit multiple classifiers for the number of labels provided

		:param X: Training matrix (with n_samples rows and n_features columns)
		:param Y: Target matrix (with n_samples rows and n_labels columns)
		:type X: matrix
		:type Y: matrix
		"""
		assert not self.fitted, "Cannot refit classifier"
		assert X.shape[0] == Y.shape[0], "The numbers of rows of X and Y should be the number of samples (and so the same)"

		self.labelCount = Y.shape[1]
		self.classifiers = [ self.classifierClass(**self.classifierArgs) for _ in range(self.labelCount) ]
		for l in range(self.labelCount):
			assert Y[:,l].min()==0, "Column %d of Y must contain negative data (class=0)"
			assert Y[:,l].max()==1, "Column %d of Y must only contain a single class (0 or 1)"

			self.classifiers[l].fit(X,Y[:,l])

		self.fitted = True

	def predict(self,X):
		"""
		Predict for multiple labels and return a matrix with predicted labels

		:param X: Testing matrix (with n_samples rows and n_features columns)
		:type X: matrix
		:return: Predicted binary matrix (with n_samples rows and n_labels columns)
		:rtype: matrix
		"""
		assert self.fitted, "Classifier has not been fit"
		
		predictions = [ self.classifiers[l].predict(X) for l in range(self.labelCount) ]

		return np.column_stack(predictions)
	

	def has_predict_proba(self):
		"""
		Returns whether the underlying classifier has the predict_proba method

		:return: Whether classifier has predict_proba method
		:rtype: bool
		"""

		return hasattr(self.classifierClass, 'predict_proba') and callable(getattr(self.classifierClass, 'predict_proba'))

	def predict_proba(self,X):
		"""
		Predict for multiple labels and return a matrix with predicted labels. Returns for the probability for the positive class (for each label column) only.

		:param X: Testing matrix (with n_samples rows and n_features columns)
		:type X: matrix
		:return: Predicted probability matrix (with n_samples rows and n_labels columns)
		:rtype: matrix
		"""
		assert self.has_predict_proba(), "Provided classifier class must have a predict_proba method"
		assert self.fitted, "Classifier has not been fit"
		
		predictions = [ self.classifiers[l].predict_proba(X)[:,1] for l in range(self.labelCount) ]

		return np.column_stack(predictions)




#!/usr/bin/env python

from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

VERSION='2.5.1'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
	long_description = f.read()

with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
	requirements = f.readlines()

setup(name='kindred',
	version=VERSION,
	description='A relation extraction toolkit for biomedical text mining',
	long_description=long_description,
	long_description_content_type='text/markdown',
	classifiers=[
		'Intended Audience :: Developers',
		'Intended Audience :: Education',
		'Intended Audience :: Information Technology',
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: MIT License',
		'Operating System :: Unix',
		'Operating System :: MacOS :: MacOS X',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3.6',
		'Topic :: Scientific/Engineering',
		'Topic :: Scientific/Engineering :: Artificial Intelligence',
		'Topic :: Scientific/Engineering :: Human Machine Interfaces',
		'Topic :: Scientific/Engineering :: Information Analysis',
		'Topic :: Text Processing',
		'Topic :: Text Processing :: General',
		'Topic :: Text Processing :: Indexing',
		'Topic :: Text Processing :: Linguistic',
	],
	url='http://github.com/jakelever/kindred',
	author='Jake Lever',
	author_email='jake.lever@gmail.com',
	license='MIT',
	packages=['kindred'],
	install_requires=requirements,
	include_package_data=True,
	zip_safe=False,
	test_suite='nose.collector',
	tests_require=['nose'])

